"""
@file
@brief Converts traces from :epkg:`nvprof`.
The source comes from `nvprof2json <https://github.com/ezyang/nvprof2json>`_.
"""

import sqlite3
import enum
import json
import copy
import io
import os
import zipfile
import cxxfilt
import pandas


def convert_trace_to_json(filename, output=None, temporary_file=None,
                          verbose=0, fLOG=None):
    """
    Converts traces produced by :epkg:`nvprof` and saved with
    format *sqlite3* (extension `.sql`). The output format
    follows `Trace Event Format
    <https://docs.google.com/document/d/
    1CvAClvFfyA5R-PhYUmn5OOQtYMH4h6I0nSsKchNAySU/preview>`_.

    :param filename: filename
    :param output: output file or None
    :param temporary_file: if the file needs to be unzipped,
        this file will be created to be the unzipped file,
        it is not cleaned after the unzipping.
    :param verbose: verbosity
    :param fLOG: logging function
    :return: json (if output is None, the list of events otherwise)

    This file, if not too big, can be viewed with `chrome-tracing`.
    The traces are usually generated by using a command line similar to:

    ::

        nvprof -o gpu_profile.sql python plot_gpu_training.py
    """
    ext = os.path.splitext(filename)[-1]
    if ext == ".zip":
        if temporary_file in (None, ''):
            temporary_file = filename + ".unzipped"
        if os.path.exists(temporary_file):
            if verbose > 0 and fLOG is not None:
                fLOG(  # pragma: no cover
                    "[convert_trace_to_json] %r already unzipped into %r"
                    "." % (filename, temporary_file))
        else:
            if verbose > 0 and fLOG is not None:
                fLOG(  # pragma: no cover
                    "[convert_trace_to_json] unzipping to file %r"
                    "." % temporary_file)
            zipf = zipfile.ZipFile(filename)
            names = zipf.namelist()
            if len(names) != 1:
                raise RuntimeError(  # pragma: no cover
                    "More than one file is stored in zip file %r." % filename)
            stream = zipf.open(names[0], "r")
            with open(temporary_file, "wb") as f:
                while True:
                    data = stream.read(65536)
                    if len(data) == 0:
                        break
                    f.write(data)
            zipf.close()
        filename = temporary_file

    conn = sqlite3.connect(filename)
    conn.row_factory = sqlite3.Row

    strings = {}
    for r in conn.execute("SELECT _id_ as id, value FROM StringTable"):
        strings[r["id"]] = _demangle(r["value"])

    traceEvents = []

    # """
    # _id_: 11625
    # cbid: 17
    # start: 1496933427584362152
    # end: 1496933427584362435
    # processId: 1317533
    # threadId: 1142654784
    # correlationId: 13119
    # returnValue: 0
    # """
    if verbose > 0 and fLOG is not None:
        fLOG("[convert_trace_to_json] step 1 begin.")
    for row in conn.execute("SELECT * FROM CUPTI_ACTIVITY_KIND_RUNTIME"):
        try:
            cbid = Cbids(row["cbid"]).name
        except ValueError:  # pragma: no cover
            cbid = str(row["cbid"])
            if verbose > 0 and fLOG is None:
                fLOG("[convert_trace_to_json] unrecognized cbid %r." % cbid)
        event = {
            "name": cbid,
            "ph": "X",  # Complete Event (Begin + End event)
            "cat": "cuda",
            "ts": _munge_time(row["start"]),
            "dur": _munge_time(row["end"] - row["start"]),
            "tid": "Thread {}: Runtime API".format(row["threadId"]),
            "pid": "[{}] Process".format(row["processId"]),
            "args": {
                # ...
            },
        }
        traceEvents.append(event)

    # DRIVER?

    # """
    # _id_: 1
    # flags: 2
    # timestamp: 1496844806028263989
    # id: 1
    # objectKind: 2
    # objectId: b'\xe5\xc0\x16\x00@\xe7\x10J\x00\x00\x00\x00'
    # name: 3
    # domain: 0
    # """
    if verbose > 0 and fLOG is not None:
        fLOG("[convert_trace_to_json] step 2 begin.")
    for row in conn.execute(" ".join([
            "SELECT",
            ",".join([
                "start.name AS name",
                "start.timestamp AS start_time",
                "end.timestamp AS end_time"
            ]),
            "FROM",
            "(SELECT * FROM CUPTI_ACTIVITY_KIND_MARKER WHERE name != 0) "
            "AS start",
            "LEFT JOIN",
            "(SELECT * FROM CUPTI_ACTIVITY_KIND_MARKER WHERE name = 0) "
            "AS end",
            "ON start.id = end.id"])):
        event = {
            "name": strings[row["name"]],
            "cat": "cuda",
            "ts": _munge_time(row["start_time"]),
            # Weirdly, these don't seem to be associated with a
            # CPU/GPU. I guess there's no CUDA Context available
            # when you run these, so it makes sense.  But nvvp
            # associates these with a GPU strangely enough
            "tid": "Markers and Ranges",
            "pid": "Markers and Ranges",
            # parse objectId?
            "args": {
                # ...
            },
        }
        if row["end_time"] is None:
            event["ph"] = "I"
        else:
            event["ph"] = "X"
            event["dur"] = _munge_time(row["end_time"] - row["start_time"])
        traceEvents.append(event)

    # """
    # _id_: 1
    # copyKind: 1
    # srcKind: 1
    # dstKind: 3
    # flags: 0
    # bytes: 7436640
    # start: 1496933426915778221
    # end: 1496933426916558424
    # deviceId: 0
    # contextId: 1
    # streamId: 7
    # correlationId: 809
    # runtimeCorrelationId: 0
    # """
    if verbose > 0 and fLOG is not None:
        fLOG("[convert_trace_to_json] step 3 begin.")
    for row in conn.execute("SELECT * FROM CUPTI_ACTIVITY_KIND_MEMCPY"):
        # copyKind:
        #   1 - Memcpy HtoD
        #   2 - Memcpy DtoH
        #   8 - Memcpy DtoD
        # flags: ???
        #   0 - Sync
        #   1 - Async
        # srcKind/dstKind
        #   1 - Pageable
        #   2 - Page-locked ???
        #   3 - Device
        # eprintRow(row)
        if row["copyKind"] == 1:
            copyKind = "HtoD"
        elif row["copyKind"] == 2:
            copyKind = "DtoH"
        elif row["copyKind"] == 8:
            copyKind = "DtoD"
        else:
            copyKind = str(row["copyKind"])
        if row["flags"] == 0:
            flags = "sync"
        elif row["flags"] == 1:
            flags = "async"
        else:
            flags = str(row["flags"])
        event = {
            "name": "Memcpy {} [{}]".format(copyKind, flags),
            "ph": "X",  # Complete Event (Begin + End event)
            "cat": "cuda",
            "ts": _munge_time(row["start"]),
            "dur": _munge_time(row["end"] - row["start"]),
            "tid": "MemCpy ({})".format(copyKind),
            # lookup GPU name.  This is tored in CUPTI_ACTIVITY_KIND_DEVICE
            "pid": "[{}:{}] Overview".format(
                row["deviceId"], row["contextId"]),
            "args": {
                "Size": _sizeof_fmt(row["bytes"]),
            },
        }
        traceEvents.append(event)

    # name: index into StringTable
    # What is thed difference between end and completed?
    # """
    # _id_: 1
    # cacheConfig: b'\x00'
    # sharedMemoryConfig: 1
    # registersPerThread: 32
    # partitionedGlobalCacheRequested: 2
    # partitionedGlobalCacheExecuted: 2
    # start: 1496844806032514222
    # end: 1496844806032531694
    # completed: 1496844806032531694
    # deviceId: 0
    # contextId: 1
    # streamId: 7
    # gridX: 57
    # gridY: 1
    # gridZ: 1
    # blockX: 128
    # blockY: 1
    # blockZ: 1
    # staticSharedMemory: 0
    # dynamicSharedMemory: 0
    # localMemoryPerThread: 0
    # localMemoryTotal: 78643200
    # correlationId: 487
    # gridId: 669
    # name: 5
    # """
    if verbose > 0 and fLOG is not None:
        fLOG("[convert_trace_to_json] step 4 begin.")
    for row in conn.execute(
            "SELECT * FROM CUPTI_ACTIVITY_KIND_CONCURRENT_KERNEL"):
        # eprint(strings[row["name"]])
        # eprintRow(row)
        event = {
            "name": strings[row["name"]],
            "ph": "X",  # Complete Event (Begin + End event)
            "cat": "cuda",
            "ts": _munge_time(row["start"]),
            "dur": _munge_time(row["end"] - row["start"]),
            "tid": "Compute",
            # lookup GPU name?
            "pid": "[{}:{}] Overview".format(
                row["deviceId"], row["contextId"]),
            "args": {
                "Grid size": "[ {}, {}, {} ]".format(
                    row["gridX"], row["gridY"], row["gridZ"]),
                "Block size": "[ {}, {}, {} ]".format(
                    row["blockX"], row["blockY"], row["blockZ"]),
                # ...
            },
        }
        alt_event = copy.deepcopy(event)
        alt_event["tid"] = alt_event["name"]
        alt_event["pid"] = "[{}:{}] Compute".format(
            row["deviceId"], row["contextId"])
        traceEvents.append(event)
        traceEvents.append(alt_event)

    if output not in (None, ''):
        if verbose > 0 and fLOG is not None:
            fLOG("[convert_trace_to_json] converting into json in %r"
                 "." % output)
        with open(output, "w", encoding="utf-8") as f:
            json.dump(traceEvents, f, separators=(',\n', ':'))
            f.write('\n')
        if verbose > 0 and fLOG is not None:
            fLOG("[convert_trace_to_json] done.")
        return traceEvents
    else:
        if verbose > 0 and fLOG is not None:
            fLOG(  # pragma: no cover
                "[convert_trace_to_json] converting into json.")
        st = io.StringIO()
        json.dump(traceEvents, st, separators=(',\n', ':'))
        st.write('\n')
        if verbose > 0 and fLOG is not None:
            fLOG("[convert_trace_to_json] done.")  # pragma: no cover
            fLOG(st.getvalue())  # pragma: no cover
        return st.getvalue()


def _munge_time(t):
    """Take a time from nvprof and convert it into a chrome://tracing time."""
    # For strict correctness, divide by 1000, but this reduces accuracy.
    return t  # / 1000.


def _demangle(name):
    """Demangle a C++ identifier using c++filt"""
    try:
        return cxxfilt.demangle(name)
    except cxxfilt.LibraryNotFound:  # pragma: no cover
        # One library is missing.
        return name


class Cbids(enum.IntEnum):
    "List of events."
    INVALID = 0
    cudaDriverGetVersion = 1
    cudaRuntimeGetVersion = 2
    cudaGetDeviceCount = 3
    cudaGetDeviceProperties = 4
    cudaChooseDevice = 5
    cudaGetChannelDesc = 6
    cudaCreateChannelDesc = 7
    cudaConfigureCall = 8
    cudaSetupArgument = 9
    cudaGetLastError = 10
    cudaPeekAtLastError = 11
    cudaGetErrorString = 12
    cudaLaunch = 13
    cudaFuncSetCacheConfig = 14
    cudaFuncGetAttributes = 15
    cudaSetDevice = 16
    cudaGetDevice = 17
    cudaSetValidDevices = 18
    cudaSetDeviceFlags = 19
    cudaMalloc = 20
    cudaMallocPitch = 21
    cudaFree = 22
    cudaMallocArray = 23
    cudaFreeArray = 24
    cudaMallocHost = 25
    cudaFreeHost = 26
    cudaHostAlloc = 27
    cudaHostGetDevicePointer = 28
    cudaHostGetFlags = 29
    cudaMemGetInfo = 30
    cudaMemcpy = 31
    cudaMemcpy2D = 32
    cudaMemcpyToArray = 33
    cudaMemcpy2DToArray = 34
    cudaMemcpyFromArray = 35
    cudaMemcpy2DFromArray = 36
    cudaMemcpyArrayToArray = 37
    cudaMemcpy2DArrayToArray = 38
    cudaMemcpyToSymbol = 39
    cudaMemcpyFromSymbol = 40
    cudaMemcpyAsync = 41
    cudaMemcpyToArrayAsync = 42
    cudaMemcpyFromArrayAsync = 43
    cudaMemcpy2DAsync = 44
    cudaMemcpy2DToArrayAsync = 45
    cudaMemcpy2DFromArrayAsync = 46
    cudaMemcpyToSymbolAsync = 47
    cudaMemcpyFromSymbolAsync = 48
    cudaMemset = 49
    cudaMemset2D = 50
    cudaMemsetAsync = 51
    cudaMemset2DAsync = 52
    cudaGetSymbolAddress = 53
    cudaGetSymbolSize = 54
    cudaBindTexture = 55
    cudaBindTexture2D = 56
    cudaBindTextureToArray = 57
    cudaUnbindTexture = 58
    cudaGetTextureAlignmentOffset = 59
    cudaGetTextureReference = 60
    cudaBindSurfaceToArray = 61
    cudaGetSurfaceReference = 62
    cudaGLSetGLDevice = 63
    cudaGLRegisterBufferObject = 64
    cudaGLMapBufferObject = 65
    cudaGLUnmapBufferObject = 66
    cudaGLUnregisterBufferObject = 67
    cudaGLSetBufferObjectMapFlags = 68
    cudaGLMapBufferObjectAsync = 69
    cudaGLUnmapBufferObjectAsync = 70
    cudaWGLGetDevice = 71
    cudaGraphicsGLRegisterImage = 72
    cudaGraphicsGLRegisterBuffer = 73
    cudaGraphicsUnregisterResource = 74
    cudaGraphicsResourceSetMapFlags = 75
    cudaGraphicsMapResources = 76
    cudaGraphicsUnmapResources = 77
    cudaGraphicsResourceGetMappedPointer = 78
    cudaGraphicsSubResourceGetMappedArray = 79
    cudaVDPAUGetDevice = 80
    cudaVDPAUSetVDPAUDevice = 81
    cudaGraphicsVDPAURegisterVideoSurface = 82
    cudaGraphicsVDPAURegisterOutputSurface = 83
    cudaD3D11GetDevice = 84
    cudaD3D11GetDevices = 85
    cudaD3D11SetDirect3DDevice = 86
    cudaGraphicsD3D11RegisterResource = 87
    cudaD3D10GetDevice = 88
    cudaD3D10GetDevices = 89
    cudaD3D10SetDirect3DDevice = 90
    cudaGraphicsD3D10RegisterResource = 91
    cudaD3D10RegisterResource = 92
    cudaD3D10UnregisterResource = 93
    cudaD3D10MapResources = 94
    cudaD3D10UnmapResources = 95
    cudaD3D10ResourceSetMapFlags = 96
    cudaD3D10ResourceGetSurfaceDimensions = 97
    cudaD3D10ResourceGetMappedArray = 98
    cudaD3D10ResourceGetMappedPointer = 99
    cudaD3D10ResourceGetMappedSize = 100
    cudaD3D10ResourceGetMappedPitch = 101
    cudaD3D9GetDevice = 102
    cudaD3D9GetDevices = 103
    cudaD3D9SetDirect3DDevice = 104
    cudaD3D9GetDirect3DDevice = 105
    cudaGraphicsD3D9RegisterResource = 106
    cudaD3D9RegisterResource = 107
    cudaD3D9UnregisterResource = 108
    cudaD3D9MapResources = 109
    cudaD3D9UnmapResources = 110
    cudaD3D9ResourceSetMapFlags = 111
    cudaD3D9ResourceGetSurfaceDimensions = 112
    cudaD3D9ResourceGetMappedArray = 113
    cudaD3D9ResourceGetMappedPointer = 114
    cudaD3D9ResourceGetMappedSize = 115
    cudaD3D9ResourceGetMappedPitch = 116
    cudaD3D9Begin = 117
    cudaD3D9End = 118
    cudaD3D9RegisterVertexBuffer = 119
    cudaD3D9UnregisterVertexBuffer = 120
    cudaD3D9MapVertexBuffer = 121
    cudaD3D9UnmapVertexBuffer = 122
    cudaThreadExit = 123
    cudaSetDoubleForDevice = 124
    cudaSetDoubleForHost = 125
    cudaThreadSynchronize = 126
    cudaThreadGetLimit = 127
    cudaThreadSetLimit = 128
    cudaStreamCreate = 129
    cudaStreamDestroy = 130
    cudaStreamSynchronize = 131
    cudaStreamQuery = 132
    cudaEventCreate = 133
    cudaEventCreateWithFlags = 134
    cudaEventRecord = 135
    cudaEventDestroy = 136
    cudaEventSynchronize = 137
    cudaEventQuery = 138
    cudaEventElapsedTime = 139
    cudaMalloc3D = 140
    cudaMalloc3DArray = 141
    cudaMemset3D = 142
    cudaMemset3DAsync = 143
    cudaMemcpy3D = 144
    cudaMemcpy3DAsync = 145
    cudaThreadSetCacheConfig = 146
    cudaStreamWaitEvent = 147
    cudaD3D11GetDirect3DDevice = 148
    cudaD3D10GetDirect3DDevice = 149
    cudaThreadGetCacheConfig = 150
    cudaPointerGetAttributes = 151
    cudaHostRegister = 152
    cudaHostUnregister = 153
    cudaDeviceCanAccessPeer = 154
    cudaDeviceEnablePeerAccess = 155
    cudaDeviceDisablePeerAccess = 156
    cudaPeerRegister = 157
    cudaPeerUnregister = 158
    cudaPeerGetDevicePointer = 159
    cudaMemcpyPeer = 160
    cudaMemcpyPeerAsync = 161
    cudaMemcpy3DPeer = 162
    cudaMemcpy3DPeerAsync = 163
    cudaDeviceReset = 164
    cudaDeviceSynchronize = 165
    cudaDeviceGetLimit = 166
    cudaDeviceSetLimit = 167
    cudaDeviceGetCacheConfig = 168
    cudaDeviceSetCacheConfig = 169
    cudaProfilerInitialize = 170
    cudaProfilerStart = 171
    cudaProfilerStop = 172
    cudaDeviceGetByPCIBusId = 173
    cudaDeviceGetPCIBusId = 174
    cudaGLGetDevices = 175
    cudaIpcGetEventHandle = 176
    cudaIpcOpenEventHandle = 177
    cudaIpcGetMemHandle = 178
    cudaIpcOpenMemHandle = 179
    cudaIpcCloseMemHandle = 180
    cudaArrayGetInfo = 181
    cudaFuncSetSharedMemConfig = 182
    cudaDeviceGetSharedMemConfig = 183
    cudaDeviceSetSharedMemConfig = 184
    cudaCreateTextureObject = 185
    cudaDestroyTextureObject = 186
    cudaGetTextureObjectResourceDesc = 187
    cudaGetTextureObjectTextureDesc = 188
    cudaCreateSurfaceObject = 189
    cudaDestroySurfaceObject = 190
    cudaGetSurfaceObjectResourceDesc = 191
    cudaMallocMipmappedArray = 192
    cudaGetMipmappedArrayLevel = 193
    cudaFreeMipmappedArray = 194
    cudaBindTextureToMipmappedArray = 195
    cudaGraphicsResourceGetMappedMipmappedArray = 196
    cudaStreamAddCallback = 197
    cudaStreamCreateWithFlags = 198
    cudaGetTextureObjectResourceViewDesc = 199
    cudaDeviceGetAttribute = 200
    cudaStreamDestroy_v5050 = 201
    cudaStreamCreateWithPriority = 202
    cudaStreamGetPriority = 203
    cudaStreamGetFlags = 204
    cudaDeviceGetStreamPriorityRange = 205
    cudaMallocManaged = 206
    cudaOccupancyMaxActiveBlocksPerMultiprocessor_v6000 = 207
    cudaStreamAttachMemAsync = 208
    cudaGetErrorName = 209
    cudaOccupancyMaxActiveBlocksPerMultiprocessor_v6050 = 210
    cudaLaunchKernel = 211
    cudaGetDeviceFlags = 212
    cudaLaunch_ptsz = 213
    cudaLaunchKernel_ptsz = 214
    cudaMemcpy_ptds = 215
    cudaMemcpy2D_ptds = 216
    cudaMemcpyToArray_ptds = 217
    cudaMemcpy2DToArray_ptds = 218
    cudaMemcpyFromArray_ptds = 219
    cudaMemcpy2DFromArray_ptds = 220
    cudaMemcpyArrayToArray_ptds = 221
    cudaMemcpy2DArrayToArray_ptds = 222
    cudaMemcpyToSymbol_ptds = 223
    cudaMemcpyFromSymbol_ptds = 224
    cudaMemcpyAsync_ptsz = 225
    cudaMemcpyToArrayAsync_ptsz = 226
    cudaMemcpyFromArrayAsync_ptsz = 227
    cudaMemcpy2DAsync_ptsz = 228
    cudaMemcpy2DToArrayAsync_ptsz = 229
    cudaMemcpy2DFromArrayAsync_ptsz = 230
    cudaMemcpyToSymbolAsync_ptsz = 231
    cudaMemcpyFromSymbolAsync_ptsz = 232
    cudaMemset_ptds = 233
    cudaMemset2D_ptds = 234
    cudaMemsetAsync_ptsz = 235
    cudaMemset2DAsync_ptsz = 236
    cudaStreamGetPriority_ptsz = 237
    cudaStreamGetFlags_ptsz = 238
    cudaStreamSynchronize_ptsz = 239
    cudaStreamQuery_ptsz = 240
    cudaStreamAttachMemAsync_ptsz = 241
    cudaEventRecord_ptsz = 242
    cudaMemset3D_ptds = 243
    cudaMemset3DAsync_ptsz = 244
    cudaMemcpy3D_ptds = 245
    cudaMemcpy3DAsync_ptsz = 246
    cudaStreamWaitEvent_ptsz = 247
    cudaStreamAddCallback_ptsz = 248
    cudaMemcpy3DPeer_ptds = 249
    cudaMemcpy3DPeerAsync_ptsz = 250
    cudaOccupancyMaxActiveBlocksPerMultiprocessorWithFlags = 251
    cudaMemPrefetchAsync = 252
    cudaMemPrefetchAsync_ptsz = 253
    cudaMemAdvise = 254
    cudaDeviceGetP2PAttribute = 255
    cudaGraphicsEGLRegisterImage = 256
    cudaEGLStreamConsumerConnect = 257
    cudaEGLStreamConsumerDisconnect = 258
    cudaEGLStreamConsumerAcquireFrame = 259
    cudaEGLStreamConsumerReleaseFrame = 260
    cudaEGLStreamProducerConnect = 261
    cudaEGLStreamProducerDisconnect = 262
    cudaEGLStreamProducerPresentFrame = 263
    cudaEGLStreamProducerReturnFrame = 264
    cudaGraphicsResourceGetMappedEglFrame = 265
    cudaMemRangeGetAttribute = 266
    cudaMemRangeGetAttributes = 267
    cudaEGLStreamConsumerConnectWithFlags = 268
    cudaLaunchCooperativeKernel = 269
    cudaLaunchCooperativeKernel_ptsz = 270
    cudaEventCreateFromEGLSync = 271
    cudaLaunchCooperativeKernelMultiDevice = 272
    cudaFuncSetAttribute = 273
    cudaImportExternalMemory = 274
    cudaExternalMemoryGetMappedBuffer = 275
    cudaExternalMemoryGetMappedMipmappedArray = 276
    cudaDestroyExternalMemory = 277
    cudaImportExternalSemaphore = 278
    cudaSignalExternalSemaphoresAsync = 279
    cudaSignalExternalSemaphoresAsync_ptsz = 280
    cudaWaitExternalSemaphoresAsync = 281
    cudaWaitExternalSemaphoresAsync_ptsz = 282
    cudaDestroyExternalSemaphore = 283
    cudaLaunchHostFunc = 284
    cudaLaunchHostFunc_ptsz = 285
    cudaGraphCreate = 286
    cudaGraphKernelNodeGetParams = 287
    cudaGraphKernelNodeSetParams = 288
    cudaGraphAddKernelNode = 289
    cudaGraphAddMemcpyNode = 290
    cudaGraphMemcpyNodeGetParams = 291
    cudaGraphMemcpyNodeSetParams = 292
    cudaGraphAddMemsetNode = 293
    cudaGraphMemsetNodeGetParams = 294
    cudaGraphMemsetNodeSetParams = 295
    cudaGraphAddHostNode = 296
    cudaGraphHostNodeGetParams = 297
    cudaGraphAddChildGraphNode = 298
    cudaGraphChildGraphNodeGetGraph = 299
    cudaGraphAddEmptyNode = 300
    cudaGraphClone = 301
    cudaGraphNodeFindInClone = 302
    cudaGraphNodeGetType = 303
    cudaGraphGetRootNodes = 304
    cudaGraphNodeGetDependencies = 305
    cudaGraphNodeGetDependentNodes = 306
    cudaGraphAddDependencies = 307
    cudaGraphRemoveDependencies = 308
    cudaGraphDestroyNode = 309
    cudaGraphInstantiate = 310
    cudaGraphLaunch = 311
    cudaGraphLaunch_ptsz = 312
    cudaGraphExecDestroy = 313
    cudaGraphDestroy = 314
    cudaStreamBeginCapture = 315
    cudaStreamBeginCapture_ptsz = 316
    cudaStreamIsCapturing = 317
    cudaStreamIsCapturing_ptsz = 318
    cudaStreamEndCapture = 319
    cudaStreamEndCapture_ptsz = 320
    cudaGraphHostNodeSetParams = 321
    cudaGraphGetNodes = 322
    cudaGraphGetEdges = 323
    cudaStreamGetCaptureInfo = 324
    cudaStreamGetCaptureInfo_ptsz = 325
    cudaGraphExecKernelNodeSetParams = 326
    cudaThreadExchangeStreamCaptureMode = 327
    cudaDeviceGetNvSciSyncAttributes = 328
    cudaOccupancyAvailableDynamicSMemPerBlock = 329
    cudaStreamSetFlags = 330
    cudaStreamSetFlags_ptsz = 331
    cudaGraphExecMemcpyNodeSetParams = 332
    cudaGraphExecMemsetNodeSetParams = 333
    cudaGraphExecHostNodeSetParams = 334
    cudaGraphExecUpdate = 335
    SIZE = 336
    FORCE_INT = 0x7FFFFFFF


def _sizeof_fmt(num, suffix='B'):
    """Format size with metric units (like nvvp)"""
    for unit in ['', 'K', 'M', 'G', 'T', 'P', 'E', 'Z']:
        if abs(num) < 1000.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1000.0  # pragma: no cover
    return "%.1f%s%s" % (num, 'Y', suffix)  # pragma: no cover


def json_to_dataframe(js):
    """
    Converts a json dump obtained with function
    @see fn convert_trace_to_json
    to a dataframe.

    :param js: a filename, a json string, a stream containing json
    :return: a dataframe
    """
    if isinstance(js, str) and os.path.exists(js):
        if len(js) < 5000:
            df = pandas.read_json(js)
        else:  # pragma: no cover
            st = io.StringIO(js)
            df = pandas.read_json(st)
    else:
        df = pandas.read_json(js)

    df['ts_sec'] = df['ts'].apply(lambda t: t / 1e9)
    return df


def json_to_dataframe_streaming(js, chunksize=100000, flatten=False, **kwargs):
    """
    Converts a big json dump (from @see fn convert_trace_to_json)
    to a dataframe. The function processes the data by streaming to avoid
    loading huge data in memory.
    Returns an iterator on dataframes.
    The function relies on :epkg:`pandas_streaming`.

    :param js: a filename, a json string, a stream containing json
    :param chunksize:
        see :func:`pandas_streaming.df.StreamingDataFrame.read_json`
    :param flatten:
        see :func:`pandas_streaming.df.StreamingDataFrame.read_json`
    :param kwargs:
        see :func:`pandas_streaming.df.StreamingDataFrame.read_json`
    :return: a dataframe
    """
    from pandas_streaming.df import StreamingDataFrame  # pylint: disable=C0415
    if isinstance(js, str):
        if len(js) < 5000 and os.path.exists(js):
            sdf = StreamingDataFrame.read_json(js)
        else:
            raise RuntimeError(
                "Use a stream or function json_to_dataframe instead of "
                "the streaming version.")
    else:
        sdf = StreamingDataFrame.read_json(js)

    sdf['ts_sec'] = sdf['ts'].apply(lambda t: t / 1e9)
    return sdf



"""
@file
@brief Onnxruntime helper.
"""
from onnxruntime.capi._pybind_state import (  # pylint: disable=E0611
    OrtDevice as C_OrtDevice, OrtValue as C_OrtValue)


def provider_to_device(provider_name):
    """
    Converts provider into a device.

    :param provider_name: provider name
    :return: device name

    .. runpython::
        :showcode:

        from onnxcustom.utils.onnxruntime_helper import provider_to_device
        print(provider_to_device('CPUExecutionProvider'))
    """
    if provider_name == 'CPUExecutionProvider':
        return 'cpu'
    if provider_name == 'CUDAExecutionProvider':
        return 'cuda'
    raise ValueError(
        "Unexpected value for provider_name=%r." % provider_name)


def get_ort_device_type(device):
    """
    Converts device into device type.

    :param device: string
    :return: device type
    """
    if isinstance(device, str):
        if device == 'cuda':
            return C_OrtDevice.cuda()
        if device == 'cpu':
            return C_OrtDevice.cpu()
        raise ValueError(  # pragma: no cover
            'Unsupported device type: %r.' % device)
    if not hasattr(device, 'device_type'):
        raise TypeError('Unsupported device type: %r.' % type(device))
    device_type = device.device_type()
    if device_type in ('cuda', 1):
        return C_OrtDevice.cuda()
    if device_type in ('cpu', 0):
        return C_OrtDevice.cpu()
    raise ValueError(  # pragma: no cover
        'Unsupported device type: %r.' % device_type)


def get_ort_device(device):
    """
    Converts device into :epkg:`C_OrtDevice`.

    :param device: any type
    :return: :epkg:`C_OrtDevice`

    Example:

    ::

        get_ort_device('cpu')
        get_ort_device('gpu')
        get_ort_device('cuda')
        get_ort_device('cuda:0')
    """
    if isinstance(device, C_OrtDevice):
        return device
    if isinstance(device, str):
        if device == 'cpu':
            return C_OrtDevice(
                C_OrtDevice.cpu(), C_OrtDevice.default_memory(), 0)
        if device in {'gpu', 'cuda:0', 'cuda', 'gpu:0'}:
            return C_OrtDevice(
                C_OrtDevice.cuda(), C_OrtDevice.default_memory(), 0)
        if device.startswith('gpu:'):
            idx = int(device[4:])
            return C_OrtDevice(
                C_OrtDevice.cuda(), C_OrtDevice.default_memory(), idx)
        if device.startswith('cuda:'):
            idx = int(device[5:])
            return C_OrtDevice(
                C_OrtDevice.cuda(), C_OrtDevice.default_memory(), idx)
        raise ValueError(
            "Unable to interpret string %r as a device." % device)
    raise TypeError(  # pragma: no cover
        "Unable to interpret type %r, (%r) as de device." % (
            type(device), device))


def ort_device_to_string(device):
    """
    Returns a string representing the device.
    Opposite of function @see fn get_ort_device.

    :param device: see :epkg:`C_OrtDevice`
    :return: string
    """
    if not isinstance(device, C_OrtDevice):
        raise TypeError(
            "device must be of type C_OrtDevice not %r." % type(device))
    ty = device.device_type()
    if ty == C_OrtDevice.cpu():
        sty = 'cpu'
    elif ty == C_OrtDevice.cuda():
        sty = 'cuda'
    else:
        raise NotImplementedError(  # pragma: no cover
            "Unable to guess device for %r and type=%r." % (device, ty))
    idx = device.device_id()
    if idx == 0:
        return sty
    return "%s:%d" % (sty, idx)


def numpy_to_ort_value(arr, device=None):
    """
    Converts a numpy array to :epkg:`C_OrtValue`.

    :param arr: numpy array
    :param device: :epkg:`C_OrtDevice` or None for cpu
    :return: :epkg:`C_OrtValue`
    """
    if device is None:
        device = get_ort_device('cpu')
    return C_OrtValue.ortvalue_from_numpy(arr, device)


def device_to_providers(device):
    """
    Returns the corresponding providers for a specific device.

    :param device: :epkg:`C_OrtDevice`
    :return: providers
    """
    if isinstance(device, str):
        device = get_ort_device(device)
    if device.device_type() == device.cpu():
        return ['CPUExecutionProvider']
    if device.device_type() == device.cuda():
        return ['CUDAExecutionProvider']
    raise ValueError(  # pragma: no cover
        "Unexpected device %r." % device)



# pylint: disable=W0105
"""
@file
@brief Helper for :epkg:`onnxruntime-training`.
"""
import inspect
from io import BytesIO
import numpy
import onnx
from onnxruntime import SessionOptions, InferenceSession, RunOptions
from onnxruntime.capi._pybind_state import (  # pylint: disable=E0611
    OrtValue as C_OrtValue)
from ..utils.onnxruntime_helper import ort_device_to_string
from .excs import ProviderError
from ._base import BaseOnnxClass


class BaseLearningOnnx(BaseOnnxClass):
    """
    Class handling ONNX function to manipulate OrtValue.
    Base class for @see cl BaseLearningRate and
    @see cl BaseLearningLoss.
    """

    def __init__(self):
        self.cache_in_ = {}
        self.cache_out_ = {}

    def __getstate__(self):
        """
        Overwrites getstate to get rid of InferenceSession.
        """
        atts = [k for k in self.__dict__ if not k.endswith('_')]
        state = {k: getattr(self, k) for k in atts}
        if hasattr(self, 'ro_'):
            state['ro_'] = True
        onx = [k for k in self.__dict__ if k.endswith('_onnx_')]
        for o in onx:
            state[o] = getattr(self, o).SerializeToString()
        onx = [k for k in self.__dict__ if k.endswith('_sess_')]
        bind = [k for k in self.__dict__ if k.endswith('_bind_')]
        for k in bind:
            state[k] = True
        binds = [k for k in self.__dict__ if k.endswith('_binds_')]
        for k in binds:
            state[k] = len(getattr(self, k))
        for o in onx:
            state[o] = getattr(self, o).get_providers()
        return state

    def __setstate__(self, state):
        """
        Overwrites getstate to get rid of InferenceSession.
        """
        for k, v in state.items():
            if k == 'ro_':
                self.ro_ = RunOptions()
            elif not k.endswith('_onnx_') and not k.endswith('_sess_'):
                setattr(self, k, v)

        so = SessionOptions()
        so.log_severity_level = 4
        for k, v in state.items():
            if k.endswith('_onnx_'):
                setattr(self, k, onnx.load(BytesIO(v)))
                k2 = k.replace("onnx", "sess")
                prov = state[k2]
                setattr(self, k2, InferenceSession(
                    getattr(self, k).SerializeToString(), so,
                    providers=prov))
        for k, v in state.items():
            if k.endswith('_bind_'):
                k2 = k[:-5]
                setattr(self, k, getattr(self, k2).io_binding()._iobinding)
            elif k.endswith('_binds_'):
                k2 = k[:-6]
                n = v
                setattr(self, k, [
                    getattr(self, k2).io_binding()._iobinding
                    for i in range(n)])
        self.cache_in_ = {}
        self.cache_out_ = {}
        return self

    def __repr_extended__(self):
        return ''

    def __repr__(self):
        """
        Usual
        """
        param = self._get_param_names()
        ps = []
        for k, v in param:
            if k not in self.__dict__:
                continue  # pragma: no cover
            ov = getattr(self, k)
            if v is not inspect._empty or ov != v:
                ro = repr(ov)
                ps.append("%s=%s" % (k, ro))
        return "%s(%s)%s" % (
            self.__class__.__name__, ", ".join(ps), self.__repr_extended__())

    def build_onnx_function(self, opset, device, *args):
        """
        This class updates the weights.
        It assumes it can do operator on *OrtValue*.
        This can be done through ONNX graph.
        This function creates :epkg:`InferenceSession`
        which do that.

        :param opset: opset to use
        :param device: :epkg:`C_OrtDevice`
        :param args: additional arguments
        """
        raise NotImplementedError(
            "This method must be overwritten.")

    @staticmethod
    def _cache_in_clear(cache, name, bind):
        key = id(bind)
        if key in cache:
            if name in cache[key]:
                if cache[key][name] == 0:
                    return True
                cache[key][name] = 0
                return False
        return True

    def clear_binding_inputs(self, name, bind, cache=False):
        """
        Clears binding and empty cache.
        """
        if cache and self._cache_in_clear(self.cache_in_, name, bind):
            return
        bind.clear_binding_inputs()

    @staticmethod
    def _bio_cache(cache, name, bind, c_ortvalue, ptr2):
        key = id(bind)
        if key in cache:
            if name in cache[key]:
                ptr = cache[key][name]
                if ptr == ptr2:
                    return True
            cache[key][name] = ptr2
        else:
            cache[key] = {name: ptr2}
        return False

    @staticmethod
    def _bio_do_bind_in(name, bind, c_ortvalue):
        bind.bind_ortvalue_input(name, c_ortvalue)

    @staticmethod
    def _bio_ptr(c):
        return c.data_ptr()

    def _bind_input_ortvalue(self, name, bind, c_ortvalue, device,
                             cache=False):
        """
        Binds :epkg:`C_OrtValue` to the structure used by
        :epkg:`InferenceSession` to run inference.

        :param name: str
        :param bind: python structure
        :param c_ortvalue: C structure for OrtValue (:epkg:`C_OrtValue`),
            it can be also a numpy array
        :param device: device
        :param cache: avoids binding again if the data pointer did not change,
            only works when c_ortvalue is of :epkg:`C_OrtValue`, the cache is
            equivalent to a dictionary
            `{ id(bind), name: c_ort_value.data_ptr() }`.
        """
        if isinstance(c_ortvalue, C_OrtValue):
            if cache and self._bio_cache(
                    self.cache_in_, name, bind, c_ortvalue,
                    self._bio_ptr(c_ortvalue)):
                return
            self._bio_do_bind_in(name, bind, c_ortvalue)
        elif isinstance(c_ortvalue, numpy.ndarray):
            if self.device_type() != device.cpu():  # pylint: disable=E1101
                raise ProviderError(
                    "device=%s is not CPU." % ort_device_to_string(
                        device))
            if cache and self._bio_cache(
                    self.cache_in_, name, bind, c_ortvalue,
                    c_ortvalue.__array_interface__['data'][0]):
                return
            bind.bind_input(
                name, device, c_ortvalue.dtype, c_ortvalue.shape,
                c_ortvalue.__array_interface__['data'][0])
        else:
            raise TypeError(  # pragma: no cover
                "Unable to bind type %r for name %r." % (
                    type(c_ortvalue), name))

    @staticmethod
    def _bio_do_bind_out(name, bind, c_ortvalue):
        bind.bind_ortvalue_output(name, c_ortvalue)

    def _bind_output_ortvalue(self, name, bind, c_ortvalue, cache=False):
        """
        Binds :epkg:`C_OrtValue` to the structure used by
        :epkg:`InferenceSession` to run inference.

        :param name: str
        :param bind: python structure
        :param c_ortvalue: C structure for OrtValue (:epkg:`C_OrtValue`)
        :param cache: avoids binding again if the data pointer did not change,
            only works when c_ortvalue is of :epkg:`C_OrtValue`, the cache is
            equivalent to a dictionary
            `{ id(bind), name: c_ort_value.data_ptr() }`.

        This method can be used for inplace computation.
        """
        if isinstance(c_ortvalue, C_OrtValue):
            if cache and self._bio_cache(
                    self.cache_out_, name, bind, c_ortvalue,
                    self._bio_ptr(c_ortvalue)):
                return
            self._bio_do_bind_out(name, bind, c_ortvalue)
        else:
            raise TypeError(  # pragma: no cover
                "Unable to bind type %r for name %r." % (
                    type(c_ortvalue), name))



# pylint: disable=C0415,E1101
"""
@file
@brief ONNX manipulations to help build ONNX gradient graphs.
"""
from collections import OrderedDict
import numpy
from onnx.mapping import TENSOR_TYPE_TO_NP_TYPE
from onnx.numpy_helper import to_array, from_array
from onnx.helper import (
    make_node, make_graph, make_model, make_tensor_value_info,
    set_model_props)
from onnx import TensorProto


def _unique_name(existing_names, name):
    """
    Returns a name different from any name in *existing_names*.

    :param existing_names: set of names
    :param name: current
    :return: unique name
    """
    if name not in existing_names:
        existing_names.add(name)
        return name
    name0 = name
    i = 2
    while name in existing_names:
        name = "%s_%d" % (name0, i)
        i += 1
    existing_names.add(name)
    return name


def _loss_l1(existing_names, elem, shape,
             output_name, label_name,
             weight_name, loss_name):
    """
    Implements loss l1.
    """
    diff_name = _unique_name(existing_names, "loss_diff")
    diff2_name = _unique_name(existing_names, "loss_diff")
    nodes = [make_node('Sub', [output_name, label_name], [diff_name]),
             make_node('Abs', [diff_name], [diff2_name])]
    if weight_name is not None:
        res_name = _unique_name(existing_names, "loss_diff_weight")
        nodes.append(
            make_node('Mul', [diff2_name, weight_name], [res_name]))
    else:
        res_name = diff2_name
    nodes.append(make_node('ReduceSum', [res_name], [loss_name]))

    inputs = [make_tensor_value_info(label_name, elem, shape)]
    if weight_name is not None:
        inputs.append(
            make_tensor_value_info(weight_name, elem, [shape[0]]))
    return (
        [], inputs, nodes,
        [make_tensor_value_info(loss_name, elem, [1, 1])])


def _loss_l2(existing_names, elem, shape,
             output_name, label_name,
             weight_name, loss_name):
    """
    Implements loss l2.
    """
    diff_name = _unique_name(existing_names, "loss_diff")
    diff2_name = _unique_name(existing_names, "loss_diff")
    nodes = [make_node('Sub', [output_name, label_name], [diff_name]),
             make_node('Mul', [diff_name, diff_name], [diff2_name])]
    if weight_name is not None:
        res_name = _unique_name(existing_names, "loss_diff_weight")
        nodes.append(
            make_node('Mul', [diff2_name, weight_name], [res_name]))
    else:
        res_name = diff2_name
    nodes.append(make_node('ReduceSum', [res_name], [loss_name]))

    inputs = [make_tensor_value_info(label_name, elem, shape)]
    if weight_name is not None:
        inputs.append(
            make_tensor_value_info(weight_name, elem, [shape[0]]))
    return (
        [], inputs, nodes,
        [make_tensor_value_info(loss_name, elem, [1, 1])])


def _loss_elastic(existing_names, elem, shape,
                  output_name, label_name,
                  weight_name, loss_name,
                  l1_weight=0.5, l2_weight=0.5):
    """
    Implements mixture of losses l1 and l2.
    """
    l1_name = _unique_name(existing_names, "l1_name")
    l2_name = _unique_name(existing_names, "l2_name")
    dtype = TENSOR_TYPE_TO_NP_TYPE[elem]
    onx_l1_weight = from_array(
        numpy.array([l1_weight], dtype=dtype), name=l1_name)
    onx_l2_weight = from_array(
        numpy.array([l2_weight], dtype=dtype), name=l2_name)
    inits = [onx_l1_weight, onx_l2_weight]

    diff_name = _unique_name(existing_names, "loss_diff")
    diff1_name = _unique_name(existing_names, "loss_l1")
    diff2_name = _unique_name(existing_names, "loss_l2")
    wl1_name = _unique_name(existing_names, "loss_l1")
    wl2_name = _unique_name(existing_names, "loss_l2")
    final_loss = _unique_name(existing_names, "final_loss")
    nodes = [make_node('Sub', [output_name, label_name], [diff_name]),
             make_node('Mul', [diff_name, diff_name], [diff2_name]),
             make_node('Abs', [diff_name], [diff1_name]),
             make_node('Mul', [diff1_name, l1_name], [wl1_name]),
             make_node('Mul', [diff2_name, l2_name], [wl2_name]),
             make_node('Add', [wl1_name, wl2_name], [final_loss]),
             ]
    if weight_name is not None:
        res_name = _unique_name(existing_names, "loss_diff_weight")
        nodes.append(
            make_node('Mul', [final_loss, weight_name], [res_name]))
    else:
        res_name = final_loss
    nodes.append(make_node('ReduceSum', [res_name], [loss_name]))

    inputs = [make_tensor_value_info(label_name, elem, shape)]
    if weight_name is not None:
        inputs.append(
            make_tensor_value_info(weight_name, elem, [shape[0]]))
    return (
        inits, inputs, nodes,
        [make_tensor_value_info(loss_name, elem, [1, 1])])


def _loss_log(existing_names, elem, shape,
              output_name, label_name,
              weight_name, loss_name,
              eps=1e-6):
    """
    This only works for a binary classification.
    The log loss is `'log(yt, yp) = (1-yt)\\log(1-yp) - yt\\log(yp)`,
    this only works for a binary classification where *yp* is the
    predicted probability, *yt* is the expected probability.
    *yt* is expected to be binary, *yp* is a matrix with two
    columns, the sum on every line is 1.
    Parameter *eps* is used to avoid computing *log(0)*.
    """
    if output_name == 'output_label':
        raise RuntimeError(  # pragma: no cover
            "output_name=%r, log loss does not work on labels."
            "" % output_name)
    dtype = TENSOR_TYPE_TO_NP_TYPE[elem]
    one_name = _unique_name(existing_names, "one_name")
    eps_name = _unique_name(existing_names, "eps_name")
    eps1_name = _unique_name(existing_names, "eps1_name")
    axes_name = _unique_name(existing_names, "axes_name")

    eps_init = from_array(numpy.array([eps], dtype=dtype), name=eps_name)
    one_init = from_array(numpy.array([1], dtype=dtype), name=one_name)
    eps1_init = from_array(
        numpy.array([1 - eps], dtype=dtype), name=eps1_name)
    axes_init = from_array(
        numpy.array([1], dtype=numpy.int64), name=axes_name)

    clip_name = _unique_name(existing_names, "clip_name")
    clip_red_name = _unique_name(existing_names, "clip_red_name")
    new_output_name = _unique_name(existing_names, "new_output_name")
    cast_name = _unique_name(existing_names, "cast_name")
    log_name = _unique_name(existing_names, "log_name")
    subl_name = _unique_name(existing_names, "subl_name")
    conc_name = _unique_name(existing_names, "conc_name")
    mul_name = _unique_name(existing_names, "mul_name")
    like_name = _unique_name(existing_names, "like_name")

    nodes = [
        make_node(
            'Clip', [output_name, eps_name, eps1_name], [clip_name]),
        make_node(
            'ReduceSum', [clip_name, axes_name], [clip_red_name], keepdims=1),
        make_node('Div', [clip_name, clip_red_name], [new_output_name]),
        make_node('Log', [new_output_name], [log_name]),
        make_node('Cast', [label_name], [cast_name], to=elem),
        make_node('Sub', [one_name, cast_name], [subl_name]),
        make_node('Concat', [subl_name, cast_name], [conc_name], axis=1),
        make_node('Mul', [log_name, conc_name], [mul_name]),
        make_node(
            'ReduceSum', [mul_name, axes_name], [like_name], keepdims=1)]

    inputs = [make_tensor_value_info(label_name, TensorProto.INT64, shape)]

    if weight_name is not None:
        inputs.append(
            make_tensor_value_info(weight_name, elem, [shape[0]]))
        likew_name = _unique_name(existing_names, "likew_name")
        nodes.append(
            make_node('Mul', [like_name, weight_name], [likew_name]))
        like_name = likew_name

    shape_name = _unique_name(existing_names, "shape_name")
    onx_shape = from_array(
        numpy.array([1, 1], dtype=numpy.int64), name=shape_name)
    reduced_loss = _unique_name(existing_names, "reduced_loss")
    neg_reduced_loss = _unique_name(existing_names, "neg_reduced_loss")
    nodes.extend([
        make_node('ReduceMean', [like_name], [reduced_loss]),
        make_node('Neg', [reduced_loss], [neg_reduced_loss]),
        make_node('Reshape', [neg_reduced_loss, shape_name], [loss_name])])

    return (
        [eps_init, eps1_init, one_init, axes_init, onx_shape],
        inputs, nodes, [make_tensor_value_info(loss_name, elem, [1, 1])])


def penalty_loss_onnx(name, dtype, l1=None, l2=None, existing_names=None):
    """
    Returns onnx nodes to compute
    :math:`|w| \\alpha + w^2 \\beta`
    where :math:`\\alpha=l1` and :math:`\\beta=l2`.

    :param name: name of weights
    :param dtype: numpy dtype
    :param l1: coefficient for L1 norm
    :param l2: coefficient for L2 norm
    :param existing_names: names already taken in the ONNX graph
    :return: initializer, nodes
    """
    suffix = name
    cst_shape = _unique_name(existing_names, "shape_%s" % suffix)
    new_name = _unique_name(existing_names, "reshaped_%s" % suffix)
    inits = [from_array(
             numpy.array([-1], dtype=numpy.int64), name=cst_shape)]
    nodes = [make_node('Reshape', [name, cst_shape], [new_name])]
    name = new_name

    if l1 is None or l1 == 0:
        if l2 is None or l2 == 0:
            raise ValueError(  # pragma: no cover
                "l1 and l2 cannot be null or None at the same time, "
                "name=%r." % name)
        l2_name = _unique_name(existing_names, "l2_weight_%s" % suffix)
        inits.extend([from_array(
            numpy.array([l2], dtype=dtype), name=l2_name)])
        mul_name = _unique_name(existing_names, "reduced0_%s" % suffix)
        red_name = _unique_name(existing_names, "reduced_%s" % suffix)
        pen_name = _unique_name(existing_names, "penalty_%s" % suffix)
        nodes.extend([
            make_node('Mul', [name, name], [mul_name]),
            make_node('ReduceSum', [mul_name], [red_name]),
            make_node('Mul', [red_name, l2_name], [pen_name])])
        return inits, nodes

    if l2 is None or l2 == 0:
        l1_name = _unique_name(existing_names, "l1_weight_%s" % suffix)
        inits.extend([from_array(
            numpy.array([l1], dtype=dtype), name=l1_name)])
        red_name = _unique_name(existing_names, "reduced_%s" % suffix)
        abs_name = _unique_name(existing_names, "absolute_%s" % suffix)
        pen_name = _unique_name(existing_names, "penalty_%s" % suffix)
        nodes.extend([
            make_node('Abs', [name], [abs_name]),
            make_node('ReduceSum', [abs_name], [red_name]),
            make_node('Mul', [red_name, l1_name], [pen_name])])
        return inits, nodes

    l1_name = _unique_name(existing_names, "l1_weight_%s" % suffix)
    l2_name = _unique_name(existing_names, "l2_weight_%s" % suffix)
    inits.extend([
        from_array(numpy.array([l1], dtype=dtype), name=l1_name),
        from_array(numpy.array([l2], dtype=dtype), name=l2_name)])

    red_name1 = _unique_name(existing_names, "reduced1_%s" % suffix)
    mul_name = _unique_name(existing_names, "reducedm_%s" % suffix)
    red_name2 = _unique_name(existing_names, "reduced2_%s" % suffix)
    abs_name = _unique_name(existing_names, "absolute_%s" % suffix)
    pen_name1 = _unique_name(existing_names, "penalty1_%s" % suffix)
    pen_name2 = _unique_name(existing_names, "penalty2_%s" % suffix)
    pen_name = _unique_name(existing_names, "penalty_%s" % suffix)
    nodes.extend([
        make_node('Mul', [name, name], [mul_name]),
        make_node('ReduceSum', [mul_name], [red_name2]),
        make_node('Mul', [red_name2, l2_name], [pen_name2]),
        make_node('Abs', [name], [abs_name]),
        make_node('ReduceSum', [abs_name], [red_name1]),
        make_node('Mul', [red_name1, l1_name], [pen_name1]),
        make_node('Add', [pen_name1, pen_name2], [pen_name])])

    return inits, nodes


def get_train_initializer(onx):
    """
    Returns the list of initializers to train.

    :return: dictionary `{name: (value, tensor)}`

    The function walk through the list of initializers and
    returns all tensors with elements from types float or double.
    """
    res = OrderedDict()
    for init in onx.graph.initializer:
        if init.data_type in (
                TensorProto.FLOAT16,  # pylint: disable=E1101
                TensorProto.FLOAT,  # pylint: disable=E1101
                TensorProto.DOUBLE):  # pylint: disable=E1101
            res[init.name] = (to_array(init), init)
    return res


def _rewrite_op_no_grad(onx):
    """
    Rewrites operators with no gradient.
    """
    set_types = set(n.op_type for n in onx.graph.node)
    if "Reciprocal" in set_types:
        from skl2onnx.algebra.onnx_ops import OnnxDiv  # pylint: disable=E0611
        from skl2onnx.common.data_types import FloatTensorType
        from .onnx_rewriter import onnx_rewrite_operator

        opset = None
        for op in onx.opset_import:
            if op.domain in ('', 'ai.onnx'):
                opset = op.version
        if opset is None:  # pragma: no cover
            from .. import get_max_opset
            opset = get_max_opset()

        node = OnnxDiv(numpy.array([1], dtype=numpy.float32),
                       'X', output_names=['Y'],
                       op_version=opset)
        rewrite_onx = node.to_onnx(
            inputs={'X': FloatTensorType()},
            outputs={'Y': FloatTensorType()},
            target_opset=opset)
        onx = onnx_rewrite_operator(onx, 'Reciprocal', rewrite_onx)

    return onx


def add_loss_output(onx, score_name='squared_error',
                    loss_name='loss', label_name='label',
                    weight_name=None, penalty=None,
                    output_index=None, **kwargs):
    """
    Modifies an ONNX graph to add operators to score and allow training.

    :param onx: onx graph
    :param score_name: name of the score
    :param loss_name: name of the output loss
    :param label_name: name of the label input
    :param weight_name: None or any value to consider weight
        while computing loss
    :param penalty: dictionary similar to the
        following one `{ weight_name: {'l1': alpha, 'l2': beta} }`
        or `{ weight_name: beta}`,
        it adds a L1 and/or L2 penalty to one input or initializer,
        penalty = :math:`|w| \\alpha + w^2 \\beta`
    :param output_index: the output used to compute the loss,
        if None, the function assumes there is only one output,
        it must be specified if there are more than 1,
        it can be an integer or a string (output name)
    :param kwargs: additional arguments for losses (see below)
    :return: modified graph

    Possible values for *score_name*:

    * `'squared_error'` or `'l2`': :math:`\\sum_i{(f(x_i)-y_i)^2}` or
      :math:`\\sum_i{w_i (f(x_i)-y_i)^2}` if *weight_name*
      is not None
    * `'absolute_error'` or `'l1`': :math:`\\sum_i{|f(x_i)-y_i|}` or
      :math:`\\sum_i{w_i |f(x_i)-y_i|}` if *weight_name*
      is not None
    * `'elastic'`: mixture of losses, kwargs must define
      *l1_weight* and *l2_weight*, undefined, default value are 0.5
    * `'log(yt, yp)'`: log loss :math:`(1-yt)\\log(1-yp) - yt\\log(yp)`,
        this only works for a binary classification where *yp* is the
        predicted probability, *yt* is the expected probability.
        *yt* is expected to be binary, *yp* is a matrix with two
        columns, the sum on every line is 1.

    See example :ref:`l-orttraining-nn-gpu`.
    Next example shows the loss with L1 and L2 loss.

    .. gdot::
        :script: DOT-SECTION

        import numpy
        from sklearn.datasets import make_regression
        from sklearn.model_selection import train_test_split
        from sklearn.linear_model import LinearRegression
        from mlprodict.onnx_conv import to_onnx
        from mlprodict.onnxrt import OnnxInference
        from onnxcustom import __max_supported_opset__ as opset
        from onnxcustom.utils.orttraining_helper import add_loss_output
        from onnxcustom.training.optimizers import OrtGradientOptimizer

        X, y = make_regression(  # pylint: disable=W0632
            100, n_features=10, bias=2, random_state=0)
        X = X.astype(numpy.float32)
        y = y.astype(numpy.float32)
        w = (numpy.random.rand(y.shape[0]) + 1).astype(X.dtype)
        X_train, _, y_train, __, w_train, ___ = train_test_split(X, y, w)
        reg = LinearRegression()
        reg.fit(X_train, y_train, sample_weight=w_train)
        reg.coef_ = reg.coef_.reshape((1, -1))
        onx = to_onnx(reg, X_train, target_opset=opset,
                      black_op={'LinearRegressor'})

        onx_loss = add_loss_output(
            onx, weight_name='weight', score_name='elastic',
            l1_weight=0.1, l2_weight=0.9)

        print("DOT-SECTION", OnnxInference(onx_loss).to_dot())

    Next example shows how to add a L2 loss with L1 and L2 penalties
    on the coefficients.

    .. gdot::
        :script: DOT-SECTION

        import numpy
        from sklearn.datasets import make_regression
        from sklearn.model_selection import train_test_split
        from sklearn.linear_model import LinearRegression
        from mlprodict.onnx_conv import to_onnx
        from mlprodict.onnxrt import OnnxInference
        from onnxcustom import __max_supported_opset__ as opset
        from onnxcustom.utils.orttraining_helper import add_loss_output
        from onnxcustom.training.optimizers import OrtGradientOptimizer

        X, y = make_regression(  # pylint: disable=W0632
            100, n_features=10, bias=2, random_state=0)
        X = X.astype(numpy.float32)
        y = y.astype(numpy.float32)
        w = (numpy.random.rand(y.shape[0]) + 1).astype(X.dtype)
        X_train, _, y_train, __, w_train, ___ = train_test_split(X, y, w)
        reg = LinearRegression()
        reg.fit(X_train, y_train, sample_weight=w_train)
        reg.coef_ = reg.coef_.reshape((1, -1))
        onx = to_onnx(reg, X_train, target_opset=opset,
                      black_op={'LinearRegressor'})

        onx_loss = add_loss_output(
            onx, weight_name='weight', score_name='elastic',
            penalty={'coef': {'l1': 0.5, 'l2':0.5},
                     'intercept': {'l1': 0.5, 'l2':0.5}})

        print("DOT-SECTION", OnnxInference(onx_loss).to_dot())
    """
    from mlprodict.onnx_tools.optim import onnx_remove_node_unused

    # rename every intermediate output call label
    def _replace(ens):
        for i in range(len(ens)):  # pylint: disable=C0200
            if ens[i] == 'label':
                ens[i] = '_label_'

    for node in onx.graph.node:
        if "_label_" in node.input or "_label_" in node.output:
            raise RuntimeError(  # pragma: no cover
                "One intermediate result contains '_label_'. "
                "It should be removed manually.\n%r" % node)
        _replace(node.input)
        _replace(node.output)

    if output_index is None:
        if len(onx.graph.output) != 1:
            raise ValueError(  # pragma: no cover
                "Unable to guess the output to compare to the "
                "expacted labels among %r." % (
                    [o.name for o in onx.graph.output]))
        outputs = onx.graph.output
        output_index = 0
    elif isinstance(output_index, int):
        outputs = [onx.graph.output[output_index]]
    elif isinstance(output_index, str):
        outputs = [(i, o) for i, o in enumerate(onx.graph.output)
                   if o.name == output_index]
        if len(outputs) != 1:
            raise ValueError(  # pragma: no cover
                "Unable to find output %r in %r." % (
                    output_index, [o.name for o in onx.graph.output]))
        output_index = outputs[0][0]
        outputs = [outputs[0][1]]
    else:
        raise TypeError(  # pragma: no cover
            "output_index must be an integer or a str not %r."
            "" % type(output_index))

    existing_names = []
    for node in onx.graph.node:
        existing_names.extend(node.output)
        existing_names.extend(node.input)
    existing_names = set(existing_names)

    output_onx = onx.graph.output[output_index]
    output_name = output_onx.name
    elem = output_onx.type.tensor_type.elem_type
    if elem == 0:
        raise TypeError(  # pragma: no cover
            "Unable to guess input tensor type from %r."
            "" % output_onx)
    shape = []
    for d in output_onx.type.tensor_type.shape.dim:
        shape.append(d.dim_value if d.dim_value > 0 else None)

    if score_name in ('squared_error', 'l2'):
        inits, inputs, nodes, outputs = _loss_l2(
            existing_names, elem, shape, output_name, label_name,
            weight_name, loss_name)
    elif score_name in ('absolute_error', 'l1'):
        inits, inputs, nodes, outputs = _loss_l1(
            existing_names, elem, shape, output_name, label_name,
            weight_name, loss_name)
    elif score_name == 'elastic':
        inits, inputs, nodes, outputs = _loss_elastic(
            existing_names, elem, shape, output_name, label_name,
            weight_name, loss_name, **kwargs)
    elif score_name == 'log':
        shape = (None, 1)
        inits, inputs, nodes, outputs = _loss_log(
            existing_names, elem, shape, output_name, label_name,
            weight_name, loss_name, **kwargs)
    else:
        raise NotImplementedError(  # pragma: no cover
            "Unexpected %r value for score_name." % score_name)

    if penalty is not None:
        final_name = nodes[-1].output[0]
        loss_name = _unique_name(existing_names, "loss_diff")
        nodes[-1].output[0] = loss_name
        names = []
        for k, v in penalty.items():
            if isinstance(v, float):
                v = {'l2': v}
            inits_to_add, nodes_to_add = penalty_loss_onnx(
                k, dtype=TENSOR_TYPE_TO_NP_TYPE[elem],
                existing_names=existing_names, **v)
            names.append(nodes_to_add[-1].output[0])
            nodes.extend(nodes_to_add)
            inits.extend(inits_to_add)
        # Operator Sum does not have a gradient.
        if len(names) == 1:
            pen_name = names[0]
        else:
            current = names[0]
            for i in range(1, len(names)):
                new_name = _unique_name(existing_names, "sumop")
                nodes.append(
                    make_node('Add', [current, names[i]], [new_name]))
                current = new_name
            pen_name = current

        cst_shape = _unique_name(existing_names, "shapevect")
        inits.append(from_array(
            numpy.array([-1, 1], dtype=numpy.int64), name=cst_shape))
        loss_reshape = _unique_name(existing_names, "loss_reshape")
        pen_reshape = _unique_name(existing_names, "penalty_reshape")
        nodes.extend([
            make_node("Reshape", [pen_name, cst_shape], [pen_reshape]),
            make_node("Reshape", [loss_name, cst_shape], [loss_reshape])])

        nodes.append(
            make_node('Add', [pen_reshape, loss_reshape], [final_name]))

    inits = list(onx.graph.initializer) + inits
    graph = make_graph(
        list(onx.graph.node) + nodes,
        onx.graph.name,
        list(onx.graph.input) + inputs,
        outputs + [onx.graph.output[output_index]],
        inits)
    onnx_model = make_model(graph)
    onnx_model.ir_version = onx.ir_version
    onnx_model.producer_name = onx.producer_name
    onnx_model.producer_version = onx.producer_version
    onnx_model.domain = onx.domain
    onnx_model.model_version = onx.model_version
    onnx_model.doc_string = onx.doc_string
    if len(onx.metadata_props) > 0:
        values = {p.key: p.value for p in onx.metadata_props}
        set_model_props(onnx_model, values)

    # fix opset import
    del onnx_model.opset_import[:]  # pylint: disable=E1101
    for oimp in onx.opset_import:
        op_set = onnx_model.opset_import.add()  # pylint: disable=E1101
        op_set.domain = oimp.domain
        op_set.version = oimp.version
    return _rewrite_op_no_grad(onnx_remove_node_unused(onnx_model))



"""
@file
@brief Informations related to the :epkg:`ImageNet` competition.
"""


def get_class_names():
    """
    Returns the class names for the :epkg:`ImageNet`
    competition as a dictionary.
    """
    return class_names


class_names = {
    0: 'tench, Tinca tinca',
    1: 'goldfish, Carassius auratus',
    2: 'great white shark, white shark, man-eater, man-eating shark, '
       'Carcharodon carcharias',
    3: 'tiger shark, Galeocerdo cuvieri',
    4: 'hammerhead, hammerhead shark',
    5: 'electric ray, crampfish, numbfish, torpedo',
    6: 'stingray',
    7: 'cock',
    8: 'hen',
    9: 'ostrich, Struthio camelus',
    10: 'brambling, Fringilla montifringilla',
    11: 'goldfinch, Carduelis carduelis',
    12: 'house finch, linnet, Carpodacus mexicanus',
    13: 'junco, snowbird',
    14: 'indigo bunting, indigo finch, indigo bird, Passerina cyanea',
    15: 'robin, American robin, Turdus migratorius',
    16: 'bulbul',
    17: 'jay',
    18: 'magpie',
    19: 'chickadee',
    20: 'water ouzel, dipper',
    21: 'kite',
    22: 'bald eagle, American eagle, Haliaeetus leucocephalus',
    23: 'vulture',
    24: 'great grey owl, great gray owl, Strix nebulosa',
    25: 'European fire salamander, Salamandra salamandra',
    26: 'common newt, Triturus vulgaris',
    27: 'eft',
    28: 'spotted salamander, Ambystoma maculatum',
    29: 'axolotl, mud puppy, Ambystoma mexicanum',
    30: 'bullfrog, Rana catesbeiana',
    31: 'tree frog, tree-frog',
    32: 'tailed frog, bell toad, ribbed toad, tailed toad, Ascaphus trui',
    33: 'loggerhead, loggerhead turtle, Caretta caretta',
    34: 'leatherback turtle, leatherback, leathery turtle, '
        'Dermochelys coriacea',
    35: 'mud turtle',
    36: 'terrapin',
    37: 'box turtle, box tortoise',
    38: 'banded gecko',
    39: 'common iguana, iguana, Iguana iguana',
    40: 'American chameleon, anole, Anolis carolinensis',
    41: 'whiptail, whiptail lizard',
    42: 'agama',
    43: 'frilled lizard, Chlamydosaurus kingi',
    44: 'alligator lizard',
    45: 'Gila monster, Heloderma suspectum',
    46: 'green lizard, Lacerta viridis',
    47: 'African chameleon, Chamaeleo chamaeleon',
    48: 'Komodo dragon, Komodo lizard, dragon lizard, giant lizard, '
        'Varanus komodoensis',
    49: 'African crocodile, Nile crocodile, Crocodylus niloticus',
    50: 'American alligator, Alligator mississipiensis',
    51: 'triceratops',
    52: 'thunder snake, worm snake, Carphophis amoenus',
    53: 'ringneck snake, ring-necked snake, ring snake',
    54: 'hognose snake, puff adder, sand viper',
    55: 'green snake, grass snake',
    56: 'king snake, kingsnake',
    57: 'garter snake, grass snake',
    58: 'water snake',
    59: 'vine snake',
    60: 'night snake, Hypsiglena torquata',
    61: 'boa constrictor, Constrictor constrictor',
    62: 'rock python, rock snake, Python sebae',
    63: 'Indian cobra, Naja naja',
    64: 'green mamba',
    65: 'sea snake',
    66: 'horned viper, cerastes, sand viper, horned asp, Cerastes cornutus',
    67: 'diamondback, diamondback rattlesnake, Crotalus adamanteus',
    68: 'sidewinder, horned rattlesnake, Crotalus cerastes',
    69: 'trilobite',
    70: 'harvestman, daddy longlegs, Phalangium opilio',
    71: 'scorpion',
    72: 'black and gold garden spider, Argiope aurantia',
    73: 'barn spider, Araneus cavaticus',
    74: 'garden spider, Aranea diademata',
    75: 'black widow, Latrodectus mactans',
    76: 'tarantula',
    77: 'wolf spider, hunting spider',
    78: 'tick',
    79: 'centipede',
    80: 'black grouse',
    81: 'ptarmigan',
    82: 'ruffed grouse, partridge, Bonasa umbellus',
    83: 'prairie chicken, prairie grouse, prairie fowl',
    84: 'peacock',
    85: 'quail',
    86: 'partridge',
    87: 'African grey, African gray, Psittacus erithacus',
    88: 'macaw',
    89: 'sulphur-crested cockatoo, Kakatoe galerita, Cacatua galerita',
    90: 'lorikeet',
    91: 'coucal',
    92: 'bee eater',
    93: 'hornbill',
    94: 'hummingbird',
    95: 'jacamar',
    96: 'toucan',
    97: 'drake',
    98: 'red-breasted merganser, Mergus serrator',
    99: 'goose',
    100: 'black swan, Cygnus atratus',
    101: 'tusker',
    102: 'echidna, spiny anteater, anteater',
    103: 'platypus, duckbill, duckbilled platypus, duck-billed '
         'platypus, Ornithorhynchus anatinus',
    104: 'wallaby, brush kangaroo',
    105: 'koala, koala bear, kangaroo bear, native bear, '
         'Phascolarctos cinereus',
    106: 'wombat',
    107: 'jellyfish',
    108: 'sea anemone, anemone',
    109: 'brain coral',
    110: 'flatworm, platyhelminth',
    111: 'nematode, nematode worm, roundworm',
    112: 'conch',
    113: 'snail',
    114: 'slug',
    115: 'sea slug, nudibranch',
    116: 'chiton, coat-of-mail shell, sea cradle, polyplacophore',
    117: 'chambered nautilus, pearly nautilus, nautilus',
    118: 'Dungeness crab, Cancer magister',
    119: 'rock crab, Cancer irroratus',
    120: 'fiddler crab',
    121: 'king crab, Alaska crab, Alaskan king crab, Alaska king '
         'crab, Paralithodes camtschatica',
    122: 'American lobster, Northern lobster, Maine lobster, '
         'Homarus americanus',
    123: 'spiny lobster, langouste, rock lobster, crawfish, '
         'crayfish, sea crawfish',
    124: 'crayfish, crawfish, crawdad, crawdaddy',
    125: 'hermit crab',
    126: 'isopod',
    127: 'white stork, Ciconia ciconia',
    128: 'black stork, Ciconia nigra',
    129: 'spoonbill',
    130: 'flamingo',
    131: 'little blue heron, Egretta caerulea',
    132: 'American egret, great white heron, Egretta albus',
    133: 'bittern',
    134: 'crane',
    135: 'limpkin, Aramus pictus',
    136: 'European gallinule, Porphyrio porphyrio',
    137: 'American coot, marsh hen, mud hen, water hen, Fulica americana',
    138: 'bustard',
    139: 'ruddy turnstone, Arenaria interpres',
    140: 'red-backed sandpiper, dunlin, Erolia alpina',
    141: 'redshank, Tringa totanus',
    142: 'dowitcher',
    143: 'oystercatcher, oyster catcher',
    144: 'pelican',
    145: 'king penguin, Aptenodytes patagonica',
    146: 'albatross, mollymawk',
    147: 'grey whale, gray whale, devilfish, Eschrichtius '
         'gibbosus, Eschrichtius robustus',
    148: 'killer whale, killer, orca, grampus, sea wolf, Orcinus orca',
    149: 'dugong, Dugong dugon',
    150: 'sea lion',
    151: 'Chihuahua',
    152: 'Japanese spaniel',
    153: 'Maltese dog, Maltese terrier, Maltese',
    154: 'Pekinese, Pekingese, Peke',
    155: 'Shih-Tzu',
    156: 'Blenheim spaniel',
    157: 'papillon',
    158: 'toy terrier',
    159: 'Rhodesian ridgeback',
    160: 'Afghan hound, Afghan',
    161: 'basset, basset hound',
    162: 'beagle',
    163: 'bloodhound, sleuthhound',
    164: 'bluetick',
    165: 'black-and-tan coonhound',
    166: 'Walker hound, Walker foxhound',
    167: 'English foxhound',
    168: 'redbone',
    169: 'borzoi, Russian wolfhound',
    170: 'Irish wolfhound',
    171: 'Italian greyhound',
    172: 'whippet',
    173: 'Ibizan hound, Ibizan Podenco',
    174: 'Norwegian elkhound, elkhound',
    175: 'otterhound, otter hound',
    176: 'Saluki, gazelle hound',
    177: 'Scottish deerhound, deerhound',
    178: 'Weimaraner',
    179: 'Staffordshire bullterrier, Staffordshire bull terrier',
    180: 'American Staffordshire terrier, Staffordshire terrier, '
         'American pit bull terrier, pit bull terrier',
    181: 'Bedlington terrier',
    182: 'Border terrier',
    183: 'Kerry blue terrier',
    184: 'Irish terrier',
    185: 'Norfolk terrier',
    186: 'Norwich terrier',
    187: 'Yorkshire terrier',
    188: 'wire-haired fox terrier',
    189: 'Lakeland terrier',
    190: 'Sealyham terrier, Sealyham',
    191: 'Airedale, Airedale terrier',
    192: 'cairn, cairn terrier',
    193: 'Australian terrier',
    194: 'Dandie Dinmont, Dandie Dinmont terrier',
    195: 'Boston bull, Boston terrier',
    196: 'miniature schnauzer',
    197: 'giant schnauzer',
    198: 'standard schnauzer',
    199: 'Scotch terrier, Scottish terrier, Scottie',
    200: 'Tibetan terrier, chrysanthemum dog',
    201: 'silky terrier, Sydney silky',
    202: 'soft-coated wheaten terrier',
    203: 'West Highland white terrier',
    204: 'Lhasa, Lhasa apso',
    205: 'flat-coated retriever',
    206: 'curly-coated retriever',
    207: 'golden retriever',
    208: 'Labrador retriever',
    209: 'Chesapeake Bay retriever',
    210: 'German short-haired pointer',
    211: 'vizsla, Hungarian pointer',
    212: 'English setter',
    213: 'Irish setter, red setter',
    214: 'Gordon setter',
    215: 'Brittany spaniel',
    216: 'clumber, clumber spaniel',
    217: 'English springer, English springer spaniel',
    218: 'Welsh springer spaniel',
    219: 'cocker spaniel, English cocker spaniel, cocker',
    220: 'Sussex spaniel',
    221: 'Irish water spaniel',
    222: 'kuvasz',
    223: 'schipperke',
    224: 'groenendael',
    225: 'malinois',
    226: 'briard',
    227: 'kelpie',
    228: 'komondor',
    229: 'Old English sheepdog, bobtail',
    230: 'Shetland sheepdog, Shetland sheep dog, Shetland',
    231: 'collie',
    232: 'Border collie',
    233: 'Bouvier des Flandres, Bouviers des Flandres',
    234: 'Rottweiler',
    235: 'German shepherd, German shepherd dog, German police dog, alsatian',
    236: 'Doberman, Doberman pinscher',
    237: 'miniature pinscher',
    238: 'Greater Swiss Mountain dog',
    239: 'Bernese mountain dog',
    240: 'Appenzeller',
    241: 'EntleBucher',
    242: 'boxer',
    243: 'bull mastiff',
    244: 'Tibetan mastiff',
    245: 'French bulldog',
    246: 'Great Dane',
    247: 'Saint Bernard, St Bernard',
    248: 'Eskimo dog, husky',
    249: 'malamute, malemute, Alaskan malamute',
    250: 'Siberian husky',
    251: 'dalmatian, coach dog, carriage dog',
    252: 'affenpinscher, monkey pinscher, monkey dog',
    253: 'basenji',
    254: 'pug, pug-dog',
    255: 'Leonberg',
    256: 'Newfoundland, Newfoundland dog',
    257: 'Great Pyrenees',
    258: 'Samoyed, Samoyede',
    259: 'Pomeranian',
    260: 'chow, chow chow',
    261: 'keeshond',
    262: 'Brabancon griffon',
    263: 'Pembroke, Pembroke Welsh corgi',
    264: 'Cardigan, Cardigan Welsh corgi',
    265: 'toy poodle',
    266: 'miniature poodle',
    267: 'standard poodle',
    268: 'Mexican hairless',
    269: 'timber wolf, grey wolf, gray wolf, Canis lupus',
    270: 'white wolf, Arctic wolf, Canis lupus tundrarum',
    271: 'red wolf, maned wolf, Canis rufus, Canis niger',
    272: 'coyote, prairie wolf, brush wolf, Canis latrans',
    273: 'dingo, warrigal, warragal, Canis dingo',
    274: 'dhole, Cuon alpinus',
    275: 'African hunting dog, hyena dog, Cape hunting dog, Lycaon pictus',
    276: 'hyena, hyaena',
    277: 'red fox, Vulpes vulpes',
    278: 'kit fox, Vulpes macrotis',
    279: 'Arctic fox, white fox, Alopex lagopus',
    280: 'grey fox, gray fox, Urocyon cinereoargenteus',
    281: 'tabby, tabby cat',
    282: 'tiger cat',
    283: 'Persian cat',
    284: 'Siamese cat, Siamese',
    285: 'Egyptian cat',
    286: 'cougar, puma, catamount, mountain lion, painter, '
         'panther, Felis concolor',
    287: 'lynx, catamount',
    288: 'leopard, Panthera pardus',
    289: 'snow leopard, ounce, Panthera uncia',
    290: 'jaguar, panther, Panthera onca, Felis onca',
    291: 'lion, king of beasts, Panthera leo',
    292: 'tiger, Panthera tigris',
    293: 'cheetah, chetah, Acinonyx jubatus',
    294: 'brown bear, bruin, Ursus arctos',
    295: 'American black bear, black bear, Ursus americanus, '
         'Euarctos americanus',
    296: 'ice bear, polar bear, Ursus Maritimus, Thalarctos maritimus',
    297: 'sloth bear, Melursus ursinus, Ursus ursinus',
    298: 'mongoose',
    299: 'meerkat, mierkat',
    300: 'tiger beetle',
    301: 'ladybug, ladybeetle, lady beetle, ladybird, ladybird beetle',
    302: 'ground beetle, carabid beetle',
    303: 'long-horned beetle, longicorn, longicorn beetle',
    304: 'leaf beetle, chrysomelid',
    305: 'dung beetle',
    306: 'rhinoceros beetle',
    307: 'weevil',
    308: 'fly',
    309: 'bee',
    310: 'ant, emmet, pismire',
    311: 'grasshopper, hopper',
    312: 'cricket',
    313: 'walking stick, walkingstick, stick insect',
    314: 'cockroach, roach',
    315: 'mantis, mantid',
    316: 'cicada, cicala',
    317: 'leafhopper',
    318: 'lacewing, lacewing fly',
    319: "dragonfly, darning needle, devil's darning needle, "
         "sewing needle, snake feeder, snake doctor, mosquito "
         "hawk, skeeter hawk",
    320: 'damselfly',
    321: 'admiral',
    322: 'ringlet, ringlet butterfly',
    323: 'monarch, monarch butterfly, milkweed butterfly, Danaus plexippus',
    324: 'cabbage butterfly',
    325: 'sulphur butterfly, sulfur butterfly',
    326: 'lycaenid, lycaenid butterfly',
    327: 'starfish, sea star',
    328: 'sea urchin',
    329: 'sea cucumber, holothurian',
    330: 'wood rabbit, cottontail, cottontail rabbit',
    331: 'hare',
    332: 'Angora, Angora rabbit',
    333: 'hamster',
    334: 'porcupine, hedgehog',
    335: 'fox squirrel, eastern fox squirrel, Sciurus niger',
    336: 'marmot',
    337: 'beaver',
    338: 'guinea pig, Cavia cobaya',
    339: 'sorrel',
    340: 'zebra',
    341: 'hog, pig, grunter, squealer, Sus scrofa',
    342: 'wild boar, boar, Sus scrofa',
    343: 'warthog',
    344: 'hippopotamus, hippo, river horse, Hippopotamus amphibius',
    345: 'ox',
    346: 'water buffalo, water ox, Asiatic buffalo, Bubalus bubalis',
    347: 'bison',
    348: 'ram, tup',
    349: 'bighorn, bighorn sheep, cimarron, Rocky Mountain '
         'bighorn, Rocky Mountain sheep, Ovis canadensis',
    350: 'ibex, Capra ibex',
    351: 'hartebeest',
    352: 'impala, Aepyceros melampus',
    353: 'gazelle',
    354: 'Arabian camel, dromedary, Camelus dromedarius',
    355: 'llama',
    356: 'weasel',
    357: 'mink',
    358: 'polecat, fitch, foulmart, foumart, Mustela putorius',
    359: 'black-footed ferret, ferret, Mustela nigripes',
    360: 'otter',
    361: 'skunk, polecat, wood pussy',
    362: 'badger',
    363: 'armadillo',
    364: 'three-toed sloth, ai, Bradypus tridactylus',
    365: 'orangutan, orang, orangutang, Pongo pygmaeus',
    366: 'gorilla, Gorilla gorilla',
    367: 'chimpanzee, chimp, Pan troglodytes',
    368: 'gibbon, Hylobates lar',
    369: 'siamang, Hylobates syndactylus, Symphalangus syndactylus',
    370: 'guenon, guenon monkey',
    371: 'patas, hussar monkey, Erythrocebus patas',
    372: 'baboon',
    373: 'macaque',
    374: 'langur',
    375: 'colobus, colobus monkey',
    376: 'proboscis monkey, Nasalis larvatus',
    377: 'marmoset',
    378: 'capuchin, ringtail, Cebus capucinus',
    379: 'howler monkey, howler',
    380: 'titi, titi monkey',
    381: 'spider monkey, Ateles geoffroyi',
    382: 'squirrel monkey, Saimiri sciureus',
    383: 'Madagascar cat, ring-tailed lemur, Lemur catta',
    384: 'indri, indris, Indri indri, Indri brevicaudatus',
    385: 'Indian elephant, Elephas maximus',
    386: 'African elephant, Loxodonta africana',
    387: 'lesser panda, red panda, panda, bear cat, cat bear, Ailurus fulgens',
    388: 'giant panda, panda, panda bear, coon bear, Ailuropoda melanoleuca',
    389: 'barracouta, snoek',
    390: 'eel',
    391: 'coho, cohoe, coho salmon, blue jack, silver salmon, '
         'Oncorhynchus kisutch',
    392: 'rock beauty, Holocanthus tricolor',
    393: 'anemone fish',
    394: 'sturgeon',
    395: 'gar, garfish, garpike, billfish, Lepisosteus osseus',
    396: 'lionfish',
    397: 'puffer, pufferfish, blowfish, globefish',
    398: 'abacus',
    399: 'abaya',
    400: "academic gown, academic robe, judge's robe",
    401: 'accordion, piano accordion, squeeze box',
    402: 'acoustic guitar',
    403: 'aircraft carrier, carrier, flattop, attack aircraft carrier',
    404: 'airliner',
    405: 'airship, dirigible',
    406: 'altar',
    407: 'ambulance',
    408: 'amphibian, amphibious vehicle',
    409: 'analog clock',
    410: 'apiary, bee house',
    411: 'apron',
    412: 'ashcan, trash can, garbage can, wastebin, ash bin, '
         'ash-bin, ashbin, dustbin, trash barrel, trash bin',
    413: 'assault rifle, assault gun',
    414: 'backpack, back pack, knapsack, packsack, rucksack, haversack',
    415: 'bakery, bakeshop, bakehouse',
    416: 'balance beam, beam',
    417: 'balloon',
    418: 'ballpoint, ballpoint pen, ballpen, Biro',
    419: 'Band Aid',
    420: 'banjo',
    421: 'bannister, banister, balustrade, balusters, handrail',
    422: 'barbell',
    423: 'barber chair',
    424: 'barbershop',
    425: 'barn',
    426: 'barometer',
    427: 'barrel, cask',
    428: 'barrow, garden cart, lawn cart, wheelbarrow',
    429: 'baseball',
    430: 'basketball',
    431: 'bassinet',
    432: 'bassoon',
    433: 'bathing cap, swimming cap',
    434: 'bath towel',
    435: 'bathtub, bathing tub, bath, tub',
    436: 'beach wagon, station wagon, wagon, estate car, '
         'beach waggon, station waggon, waggon',
    437: 'beacon, lighthouse, beacon light, pharos',
    438: 'beaker',
    439: 'bearskin, busby, shako',
    440: 'beer bottle',
    441: 'beer glass',
    442: 'bell cote, bell cot',
    443: 'bib',
    444: 'bicycle-built-for-two, tandem bicycle, tandem',
    445: 'bikini, two-piece',
    446: 'binder, ring-binder',
    447: 'binoculars, field glasses, opera glasses',
    448: 'birdhouse',
    449: 'boathouse',
    450: 'bobsled, bobsleigh, bob',
    451: 'bolo tie, bolo, bola tie, bola',
    452: 'bonnet, poke bonnet',
    453: 'bookcase',
    454: 'bookshop, bookstore, bookstall',
    455: 'bottlecap',
    456: 'bow',
    457: 'bow tie, bow-tie, bowtie',
    458: 'brass, memorial tablet, plaque',
    459: 'brassiere, bra, bandeau',
    460: 'breakwater, groin, groyne, mole, bulwark, seawall, jetty',
    461: 'breastplate, aegis, egis',
    462: 'broom',
    463: 'bucket, pail',
    464: 'buckle',
    465: 'bulletproof vest',
    466: 'bullet train, bullet',
    467: 'butcher shop, meat market',
    468: 'cab, hack, taxi, taxicab',
    469: 'caldron, cauldron',
    470: 'candle, taper, wax light',
    471: 'cannon',
    472: 'canoe',
    473: 'can opener, tin opener',
    474: 'cardigan',
    475: 'car mirror',
    476: 'carousel, carrousel, merry-go-round, roundabout, whirligig',
    477: "carpenter's kit, tool kit",
    478: 'carton',
    479: 'car wheel',
    480: 'cash machine, cash dispenser, automated teller '
         'machine, automatic teller machine, automated teller, '
         'automatic teller, ATM',
    481: 'cassette',
    482: 'cassette player',
    483: 'castle',
    484: 'catamaran',
    485: 'CD player',
    486: 'cello, violoncello',
    487: 'cellular telephone, cellular phone, cellphone, cell, mobile phone',
    488: 'chain',
    489: 'chainlink fence',
    490: 'chain mail, ring mail, mail, chain armor, chain '
         'armour, ring armor, ring armour',
    491: 'chain saw, chainsaw',
    492: 'chest',
    493: 'chiffonier, commode',
    494: 'chime, bell, gong',
    495: 'china cabinet, china closet',
    496: 'Christmas stocking',
    497: 'church, church building',
    498: 'cinema, movie theater, movie theatre, movie house, picture palace',
    499: 'cleaver, meat cleaver, chopper',
    500: 'cliff dwelling',
    501: 'cloak',
    502: 'clog, geta, patten, sabot',
    503: 'cocktail shaker',
    504: 'coffee mug',
    505: 'coffeepot',
    506: 'coil, spiral, volute, whorl, helix',
    507: 'combination lock',
    508: 'computer keyboard, keypad',
    509: 'confectionery, confectionary, candy store',
    510: 'container ship, containership, container vessel',
    511: 'convertible',
    512: 'corkscrew, bottle screw',
    513: 'cornet, horn, trumpet, trump',
    514: 'cowboy boot',
    515: 'cowboy hat, ten-gallon hat',
    516: 'cradle',
    517: 'crane',
    518: 'crash helmet',
    519: 'crate',
    520: 'crib, cot',
    521: 'Crock Pot',
    522: 'croquet ball',
    523: 'crutch',
    524: 'cuirass',
    525: 'dam, dike, dyke',
    526: 'desk',
    527: 'desktop computer',
    528: 'dial telephone, dial phone',
    529: 'diaper, nappy, napkin',
    530: 'digital clock',
    531: 'digital watch',
    532: 'dining table, board',
    533: 'dishrag, dishcloth',
    534: 'dishwasher, dish washer, dishwashing machine',
    535: 'disk brake, disc brake',
    536: 'dock, dockage, docking facility',
    537: 'dogsled, dog sled, dog sleigh',
    538: 'dome',
    539: 'doormat, welcome mat',
    540: 'drilling platform, offshore rig',
    541: 'drum, membranophone, tympan',
    542: 'drumstick',
    543: 'dumbbell',
    544: 'Dutch oven',
    545: 'electric fan, blower',
    546: 'electric guitar',
    547: 'electric locomotive',
    548: 'entertainment center',
    549: 'envelope',
    550: 'espresso maker',
    551: 'face powder',
    552: 'feather boa, boa',
    553: 'file, file cabinet, filing cabinet',
    554: 'fireboat',
    555: 'fire engine, fire truck',
    556: 'fire screen, fireguard',
    557: 'flagpole, flagstaff',
    558: 'flute, transverse flute',
    559: 'folding chair',
    560: 'football helmet',
    561: 'forklift',
    562: 'fountain',
    563: 'fountain pen',
    564: 'four-poster',
    565: 'freight car',
    566: 'French horn, horn',
    567: 'frying pan, frypan, skillet',
    568: 'fur coat',
    569: 'garbage truck, dustcart',
    570: 'gasmask, respirator, gas helmet',
    571: 'gas pump, gasoline pump, petrol pump, island dispenser',
    572: 'goblet',
    573: 'go-kart',
    574: 'golf ball',
    575: 'golfcart, golf cart',
    576: 'gondola',
    577: 'gong, tam-tam',
    578: 'gown',
    579: 'grand piano, grand',
    580: 'greenhouse, nursery, glasshouse',
    581: 'grille, radiator grille',
    582: 'grocery store, grocery, food market, market',
    583: 'guillotine',
    584: 'hair slide',
    585: 'hair spray',
    586: 'half track',
    587: 'hammer',
    588: 'hamper',
    589: 'hand blower, blow dryer, blow drier, hair dryer, hair drier',
    590: 'hand-held computer, hand-held microcomputer',
    591: 'handkerchief, hankie, hanky, hankey',
    592: 'hard disc, hard disk, fixed disk',
    593: 'harmonica, mouth organ, harp, mouth harp',
    594: 'harp',
    595: 'harvester, reaper',
    596: 'hatchet',
    597: 'holster',
    598: 'home theater, home theatre',
    599: 'honeycomb',
    600: 'hook, claw',
    601: 'hoopskirt, crinoline',
    602: 'horizontal bar, high bar',
    603: 'horse cart, horse-cart',
    604: 'hourglass',
    605: 'iPod',
    606: 'iron, smoothing iron',
    607: "jack-o'-lantern",
    608: 'jean, blue jean, denim',
    609: 'jeep, landrover',
    610: 'jersey, T-shirt, tee shirt',
    611: 'jigsaw puzzle',
    612: 'jinrikisha, ricksha, rickshaw',
    613: 'joystick',
    614: 'kimono',
    615: 'knee pad',
    616: 'knot',
    617: 'lab coat, laboratory coat',
    618: 'ladle',
    619: 'lampshade, lamp shade',
    620: 'laptop, laptop computer',
    621: 'lawn mower, mower',
    622: 'lens cap, lens cover',
    623: 'letter opener, paper knife, paperknife',
    624: 'library',
    625: 'lifeboat',
    626: 'lighter, light, igniter, ignitor',
    627: 'limousine, limo',
    628: 'liner, ocean liner',
    629: 'lipstick, lip rouge',
    630: 'Loafer',
    631: 'lotion',
    632: 'loudspeaker, speaker, speaker unit, loudspeaker '
         'system, speaker system',
    633: "loupe, jeweler's loupe",
    634: 'lumbermill, sawmill',
    635: 'magnetic compass',
    636: 'mailbag, postbag',
    637: 'mailbox, letter box',
    638: 'maillot',
    639: 'maillot, tank suit',
    640: 'manhole cover',
    641: 'maraca',
    642: 'marimba, xylophone',
    643: 'mask',
    644: 'matchstick',
    645: 'maypole',
    646: 'maze, labyrinth',
    647: 'measuring cup',
    648: 'medicine chest, medicine cabinet',
    649: 'megalith, megalithic structure',
    650: 'microphone, mike',
    651: 'microwave, microwave oven',
    652: 'military uniform',
    653: 'milk can',
    654: 'minibus',
    655: 'miniskirt, mini',
    656: 'minivan',
    657: 'missile',
    658: 'mitten',
    659: 'mixing bowl',
    660: 'mobile home, manufactured home',
    661: 'Model T',
    662: 'modem',
    663: 'monastery',
    664: 'monitor',
    665: 'moped',
    666: 'mortar',
    667: 'mortarboard',
    668: 'mosque',
    669: 'mosquito net',
    670: 'motor scooter, scooter',
    671: 'mountain bike, all-terrain bike, off-roader',
    672: 'mountain tent',
    673: 'mouse, computer mouse',
    674: 'mousetrap',
    675: 'moving van',
    676: 'muzzle',
    677: 'nail',
    678: 'neck brace',
    679: 'necklace',
    680: 'nipple',
    681: 'notebook, notebook computer',
    682: 'obelisk',
    683: 'oboe, hautboy, hautbois',
    684: 'ocarina, sweet potato',
    685: 'odometer, hodometer, mileometer, milometer',
    686: 'oil filter',
    687: 'organ, pipe organ',
    688: 'oscilloscope, scope, cathode-ray oscilloscope, CRO',
    689: 'overskirt',
    690: 'oxcart',
    691: 'oxygen mask',
    692: 'packet',
    693: 'paddle, boat paddle',
    694: 'paddlewheel, paddle wheel',
    695: 'padlock',
    696: 'paintbrush',
    697: "pajama, pyjama, pj's, jammies",
    698: 'palace',
    699: 'panpipe, pandean pipe, syrinx',
    700: 'paper towel',
    701: 'parachute, chute',
    702: 'parallel bars, bars',
    703: 'park bench',
    704: 'parking meter',
    705: 'passenger car, coach, carriage',
    706: 'patio, terrace',
    707: 'pay-phone, pay-station',
    708: 'pedestal, plinth, footstall',
    709: 'pencil box, pencil case',
    710: 'pencil sharpener',
    711: 'perfume, essence',
    712: 'Petri dish',
    713: 'photocopier',
    714: 'pick, plectrum, plectron',
    715: 'pickelhaube',
    716: 'picket fence, paling',
    717: 'pickup, pickup truck',
    718: 'pier',
    719: 'piggy bank, penny bank',
    720: 'pill bottle',
    721: 'pillow',
    722: 'ping-pong ball',
    723: 'pinwheel',
    724: 'pirate, pirate ship',
    725: 'pitcher, ewer',
    726: "plane, carpenter's plane, woodworking plane",
    727: 'planetarium',
    728: 'plastic bag',
    729: 'plate rack',
    730: 'plow, plough',
    731: "plunger, plumber's helper",
    732: 'Polaroid camera, Polaroid Land camera',
    733: 'pole',
    734: 'police van, police wagon, paddy wagon, patrol wagon, '
         'wagon, black Maria',
    735: 'poncho',
    736: 'pool table, billiard table, snooker table',
    737: 'pop bottle, soda bottle',
    738: 'pot, flowerpot',
    739: "potter's wheel",
    740: 'power drill',
    741: 'prayer rug, prayer mat',
    742: 'printer',
    743: 'prison, prison house',
    744: 'projectile, missile',
    745: 'projector',
    746: 'puck, hockey puck',
    747: 'punching bag, punch bag, punching ball, punchball',
    748: 'purse',
    749: 'quill, quill pen',
    750: 'quilt, comforter, comfort, puff',
    751: 'racer, race car, racing car',
    752: 'racket, racquet',
    753: 'radiator',
    754: 'radio, wireless',
    755: 'radio telescope, radio reflector',
    756: 'rain barrel',
    757: 'recreational vehicle, RV, R.V.',
    758: 'reel',
    759: 'reflex camera',
    760: 'refrigerator, icebox',
    761: 'remote control, remote',
    762: 'restaurant, eating house, eating place, eatery',
    763: 'revolver, six-gun, six-shooter',
    764: 'rifle',
    765: 'rocking chair, rocker',
    766: 'rotisserie',
    767: 'rubber eraser, rubber, pencil eraser',
    768: 'rugby ball',
    769: 'rule, ruler',
    770: 'running shoe',
    771: 'safe',
    772: 'safety pin',
    773: 'saltshaker, salt shaker',
    774: 'sandal',
    775: 'sarong',
    776: 'sax, saxophone',
    777: 'scabbard',
    778: 'scale, weighing machine',
    779: 'school bus',
    780: 'schooner',
    781: 'scoreboard',
    782: 'screen, CRT screen',
    783: 'screw',
    784: 'screwdriver',
    785: 'seat belt, seatbelt',
    786: 'sewing machine',
    787: 'shield, buckler',
    788: 'shoe shop, shoe-shop, shoe store',
    789: 'shoji',
    790: 'shopping basket',
    791: 'shopping cart',
    792: 'shovel',
    793: 'shower cap',
    794: 'shower curtain',
    795: 'ski',
    796: 'ski mask',
    797: 'sleeping bag',
    798: 'slide rule, slipstick',
    799: 'sliding door',
    800: 'slot, one-armed bandit',
    801: 'snorkel',
    802: 'snowmobile',
    803: 'snowplow, snowplough',
    804: 'soap dispenser',
    805: 'soccer ball',
    806: 'sock',
    807: 'solar dish, solar collector, solar furnace',
    808: 'sombrero',
    809: 'soup bowl',
    810: 'space bar',
    811: 'space heater',
    812: 'space shuttle',
    813: 'spatula',
    814: 'speedboat',
    815: "spider web, spider's web",
    816: 'spindle',
    817: 'sports car, sport car',
    818: 'spotlight, spot',
    819: 'stage',
    820: 'steam locomotive',
    821: 'steel arch bridge',
    822: 'steel drum',
    823: 'stethoscope',
    824: 'stole',
    825: 'stone wall',
    826: 'stopwatch, stop watch',
    827: 'stove',
    828: 'strainer',
    829: 'streetcar, tram, tramcar, trolley, trolley car',
    830: 'stretcher',
    831: 'studio couch, day bed',
    832: 'stupa, tope',
    833: 'submarine, pigboat, sub, U-boat',
    834: 'suit, suit of clothes',
    835: 'sundial',
    836: 'sunglass',
    837: 'sunglasses, dark glasses, shades',
    838: 'sunscreen, sunblock, sun blocker',
    839: 'suspension bridge',
    840: 'swab, swob, mop',
    841: 'sweatshirt',
    842: 'swimming trunks, bathing trunks',
    843: 'swing',
    844: 'switch, electric switch, electrical switch',
    845: 'syringe',
    846: 'table lamp',
    847: 'tank, army tank, armored combat vehicle, armoured combat vehicle',
    848: 'tape player',
    849: 'teapot',
    850: 'teddy, teddy bear',
    851: 'television, television system',
    852: 'tennis ball',
    853: 'thatch, thatched roof',
    854: 'theater curtain, theatre curtain',
    855: 'thimble',
    856: 'thresher, thrasher, threshing machine',
    857: 'throne',
    858: 'tile roof',
    859: 'toaster',
    860: 'tobacco shop, tobacconist shop, tobacconist',
    861: 'toilet seat',
    862: 'torch',
    863: 'totem pole',
    864: 'tow truck, tow car, wrecker',
    865: 'toyshop',
    866: 'tractor',
    867: 'trailer truck, tractor trailer, trucking rig, rig, '
         'articulated lorry, semi',
    868: 'tray',
    869: 'trench coat',
    870: 'tricycle, trike, velocipede',
    871: 'trimaran',
    872: 'tripod',
    873: 'triumphal arch',
    874: 'trolleybus, trolley coach, trackless trolley',
    875: 'trombone',
    876: 'tub, vat',
    877: 'turnstile',
    878: 'typewriter keyboard',
    879: 'umbrella',
    880: 'unicycle, monocycle',
    881: 'upright, upright piano',
    882: 'vacuum, vacuum cleaner',
    883: 'vase',
    884: 'vault',
    885: 'velvet',
    886: 'vending machine',
    887: 'vestment',
    888: 'viaduct',
    889: 'violin, fiddle',
    890: 'volleyball',
    891: 'waffle iron',
    892: 'wall clock',
    893: 'wallet, billfold, notecase, pocketbook',
    894: 'wardrobe, closet, press',
    895: 'warplane, military plane',
    896: 'washbasin, handbasin, washbowl, lavabo, wash-hand basin',
    897: 'washer, automatic washer, washing machine',
    898: 'water bottle',
    899: 'water jug',
    900: 'water tower',
    901: 'whiskey jug',
    902: 'whistle',
    903: 'wig',
    904: 'window screen',
    905: 'window shade',
    906: 'Windsor tie',
    907: 'wine bottle',
    908: 'wing',
    909: 'wok',
    910: 'wooden spoon',
    911: 'wool, woolen, woollen',
    912: 'worm fence, snake fence, snake-rail fence, Virginia fence',
    913: 'wreck',
    914: 'yawl',
    915: 'yurt',
    916: 'web site, website, internet site, site',
    917: 'comic book',
    918: 'crossword puzzle, crossword',
    919: 'street sign',
    920: 'traffic light, traffic signal, stoplight',
    921: 'book jacket, dust cover, dust jacket, dust wrapper',
    922: 'menu',
    923: 'plate',
    924: 'guacamole',
    925: 'consomme',
    926: 'hot pot, hotpot',
    927: 'trifle',
    928: 'ice cream, icecream',
    929: 'ice lolly, lolly, lollipop, popsicle',
    930: 'French loaf',
    931: 'bagel, beigel',
    932: 'pretzel',
    933: 'cheeseburger',
    934: 'hotdog, hot dog, red hot',
    935: 'mashed potato',
    936: 'head cabbage',
    937: 'broccoli',
    938: 'cauliflower',
    939: 'zucchini, courgette',
    940: 'spaghetti squash',
    941: 'acorn squash',
    942: 'butternut squash',
    943: 'cucumber, cuke',
    944: 'artichoke, globe artichoke',
    945: 'bell pepper',
    946: 'cardoon',
    947: 'mushroom',
    948: 'Granny Smith',
    949: 'strawberry',
    950: 'orange',
    951: 'lemon',
    952: 'fig',
    953: 'pineapple, ananas',
    954: 'banana',
    955: 'jackfruit, jak, jack',
    956: 'custard apple',
    957: 'pomegranate',
    958: 'hay',
    959: 'carbonara',
    960: 'chocolate sauce, chocolate syrup',
    961: 'dough',
    962: 'meat loaf, meatloaf',
    963: 'pizza, pizza pie',
    964: 'potpie',
    965: 'burrito',
    966: 'red wine',
    967: 'espresso',
    968: 'cup',
    969: 'eggnog',
    970: 'alp',
    971: 'bubble',
    972: 'cliff, drop, drop-off',
    973: 'coral reef',
    974: 'geyser',
    975: 'lakeside, lakeshore',
    976: 'promontory, headland, head, foreland',
    977: 'sandbar, sand bar',
    978: 'seashore, coast, seacoast, sea-coast',
    979: 'valley, vale',
    980: 'volcano',
    981: 'ballplayer, baseball player',
    982: 'groom, bridegroom',
    983: 'scuba diver',
    984: 'rapeseed',
    985: 'daisy',
    986: "yellow lady's slipper, yellow lady-slipper, Cypripedium "
         "calceolus, Cypripedium parviflorum",
    987: 'corn',
    988: 'acorn',
    989: 'hip, rose hip, rosehip',
    990: 'buckeye, horse chestnut, conker',
    991: 'coral fungus',
    992: 'agaric',
    993: 'gyromitra',
    994: 'stinkhorn, carrion fungus',
    995: 'earthstar',
    996: 'hen-of-the-woods, hen of the woods, Polyporus frondosus, '
         'Grifola frondosa',
    997: 'bolete',
    998: 'ear, spike, capitulum',
    999: 'toilet tissue, toilet paper, bathroom tissue'}



# pylint: disable=C0415,E0611,E1101
"""
@file
@brief Onnx implementation of common functions used to train a model.
"""
import math
import numpy
from onnx import TensorProto, numpy_helper, helper


def onnx_rename_weights(onx):
    """
    Renames ONNX initializers to make sure their name
    follows the alphabetical order. The model is
    modified inplace. This function calls
    :func:`onnx_rename_names
    <mlprodict.onnx_tools.onnx_manipulations.onnx_rename_names>`.

    :param onx: ONNX model
    :return: same model

    .. note::
        The function does not go into subgraphs.
    """
    from mlprodict.onnx_tools.onnx_manipulations import (  # pylint: disable=C0415
        onnx_rename_names)

    init = [init.name for init in onx.graph.initializer]
    ninit = max(1, int(math.log(len(init)) / math.log(10) + 1))
    fmt = "I%0{}d_%s".format(ninit)
    new_names = [fmt % (i, name) for i, name in enumerate(init)]
    repl = dict(zip(init, new_names))
    return onnx_rename_names(onx, recursive=False, replace=repl)


def get_onnx_opset(onx, domain=''):
    """
    Returns the opset associated to an opset.

    :param onx: onx graph
    :param domain: domain
    :return: value
    """
    for opset in onx.opset_import:
        if opset.domain == domain:
            return opset.version
    raise ValueError(
        "Unable to find opset for domain=%r." % domain)


def proto_type_to_dtype(proto_type):
    """
    Converts a ONNX TensorProto type into numpy type.

    :param proto_type: integer
    :return: proto type
    """
    if proto_type == TensorProto.FLOAT:
        return numpy.float32
    if proto_type == TensorProto.DOUBLE:
        return numpy.float64
    # Not efficient.
    if proto_type == 'tensor(float)':
        return numpy.float32
    if proto_type == 'tensor(double)':
        return numpy.float64
    raise ValueError(
        "Unexpected value proto_type=%r (type=%r)." % (
            proto_type, type(proto_type)))


def dtype_to_var_type(dtype):
    """
    Converts a numpy dtype into a var type.
    """
    from skl2onnx.common.data_types import (
        FloatTensorType, DoubleTensorType,
        Int32TensorType, Int64TensorType)
    if dtype == numpy.float32:
        return FloatTensorType
    if dtype == numpy.float64:
        return DoubleTensorType
    if dtype == numpy.int64:
        return Int64TensorType
    if dtype == numpy.int32:
        return Int32TensorType
    raise ValueError(
        "Unexpected value dtype=%r." % dtype)


def add_initializer(model, name, value):
    """
    Adds an initializer to graph.

    :param model: onnx model
    :param name: initializer name
    :param value: value
    :return: new ONNX graph
    """
    inits = set(i.name for i in model.graph.initializer)
    if name in inits:
        raise ValueError(  # pragma: no cover
            "Name %r is already taken among %r." % (
                name, inits))
    list_inits = list(model.graph.initializer)
    list_inits.append(
        numpy_helper.from_array(value, name=name))
    graph_def = helper.make_graph(
        model.graph.node, model.graph.name,
        model.graph.input, model.graph.output,
        list_inits)
    onnx_model = helper.make_model(graph_def)
    onnx_model.ir_version = model.ir_version
    onnx_model.producer_name = model.producer_name
    onnx_model.producer_version = model.producer_version
    onnx_model.domain = model.domain
    onnx_model.model_version = model.model_version
    onnx_model.doc_string = model.doc_string
    if len(model.metadata_props) > 0:  # pragma: no cover
        values = {p.key: p.value for p in model.metadata_props}
        helper.set_model_props(onnx_model, values)

    del onnx_model.opset_import[:]  # pylint: disable=E1101
    for oimp in model.opset_import:
        op_set = onnx_model.opset_import.add()  # pylint: disable=E1101
        op_set.domain = oimp.domain
        op_set.version = oimp.version
    return onnx_model



# pylint: disable=W0105
"""
@file
@brief Helper for :epkg:`onnxruntime-training`.
"""
from onnxruntime import SessionOptions, InferenceSession, RunOptions
from ..utils.onnx_function import function_onnx_graph
from ..utils.onnxruntime_helper import device_to_providers
from ._base_onnx_function import BaseLearningOnnx


class BaseLearningLoss(BaseLearningOnnx):
    """
    Class handling the loss for class
    @see cl OrtGradientForwardBackwardOptimizer.
    All classes inheriting from this one creates one ONNX function,
    returning the loss and the gradient of the loss against the
    outputs. Method `loss_gradient` is the main method, it computes
    the loss and the gradient defiend by one ONNX graph and
    executed by an instance of :epkg:`InferenceSession`.
    """

    def __init__(self):
        BaseLearningOnnx.__init__(self)
        self.ro_ = RunOptions()

    def _call_iobinding(self, sess, bind):
        sess.run_with_iobinding(bind, self.ro_)

    def loss_gradient(  # pylint: disable=E1101
            self, device, expected, predicted, weight=None):
        """
        Returns the loss and the gradient as OrtValue.

        :param device: device where the training takes place
        :param expected: expected value
        :param predicted: predicted value
        :param weight: optional, training weights
            (same dimension as expected and predicted tensors)
        :return: loss and gradient
        """
        if (not hasattr(self, "loss_grad_sess_") or
                not hasattr(self, "loss_grad_sess_bind_")):
            raise RuntimeError(  # pragma: no cover
                "Attributes 'loss_grad_sess_bind_' or 'loss_grad_sess_' is "
                "missing. Method 'build_onnx_function' has not been called.")
        bind = self.loss_grad_sess_bind_
        if weight is not None:
            self._bind_input_ortvalue(
                "weight", bind, weight, device, cache=True)
        else:
            self.clear_binding_inputs("weight", bind, cache=True)
        self._bind_input_ortvalue("X1", bind, expected, device, cache=True)
        self._bind_input_ortvalue("X2", bind, predicted, device, cache=True)
        self.loss_grad_sess_bind_.bind_output('Y', device)
        self.loss_grad_sess_bind_.bind_output('Z', device)
        self._call_iobinding(self.loss_grad_sess_._sess, bind)
        loss, grad = bind.get_outputs()
        return loss, grad

    @staticmethod
    def select(class_name, **kwargs):
        """
        Returns an instance of a given initialized with
        *kwargs*.
        :param class_name: an instance of @see cl BaseLearningLoss
            or a string among the following class names (see below)
        :return: instance of @see cl BaseLearningLoss

        Possible values for *class_name*:
        * `'square_error'`: see @see cl SquareLearningLoss
        * `'absolute_error'`: see @see cl AbsoluteLearningLoss
        * `'elastic_error'`: see @see cl ElasticLearningLoss
        """
        if isinstance(class_name, BaseLearningLoss):
            return class_name
        cls = {SquareLearningLoss: ['square_error', 'square'],
               AbsoluteLearningLoss: ['absolute_error', 'absolute'],
               ElasticLearningLoss: ['elastic_error', 'elastic'],
               NegLogLearningLoss: ['log', 'neglog', 'logloss']}
        for cl, aliases in cls.items():
            if class_name == cl.__class__.__name__ or class_name in aliases:
                return cl(**kwargs)
        raise ValueError(  # pragma: no cover
            "Unexpected class name %r. It should be one of %r." % (
                class_name, list(map(lambda c: c.__name__, cls))))


class SquareLearningLoss(BaseLearningLoss):
    """
    Implements a square loss :math:`(Y - Z)^2`
    where *Y* is the output and *Z* the expected output.
    See @see fn _onnx_grad_loss_square_error for the ONNX
    implementation.
    """

    def __init__(self):
        BaseLearningLoss.__init__(self)

    def build_onnx_function(self, opset, device, weight_name):
        so = SessionOptions()
        so.log_severity_level = 4

        # loss_grad
        self.loss_grad_onnx_ = function_onnx_graph(
            "grad_loss_square_error", target_opset=opset,
            weight_name=weight_name, multiply=1)
        self.loss_grad_sess_ = InferenceSession(
            self.loss_grad_onnx_.SerializeToString(), so,
            providers=device_to_providers(device))
        self.loss_grad_sess_bind_ = (
            self.loss_grad_sess_.io_binding()._iobinding)


class AbsoluteLearningLoss(BaseLearningLoss):
    """
    Implements a square loss :math:`|Y - Z|`
    where *Y* is the output and *Z* the expected output.
    See @see fn _onnx_grad_loss_absolute_error for the ONNX
    implementation.
    """

    def __init__(self):
        BaseLearningLoss.__init__(self)

    def build_onnx_function(self, opset, device, weight_name):
        so = SessionOptions()
        so.log_severity_level = 4

        # loss_grad
        self.loss_grad_onnx_ = function_onnx_graph(
            "grad_loss_absolute_error", target_opset=opset,
            weight_name=weight_name)
        self.loss_grad_sess_ = InferenceSession(
            self.loss_grad_onnx_.SerializeToString(), so,
            providers=device_to_providers(device))
        self.loss_grad_sess_bind_ = (
            self.loss_grad_sess_.io_binding()._iobinding)


class ElasticLearningLoss(BaseLearningLoss):
    """
    Implements a square loss
    :math:`(Y - Z)^2 \\alpha + |Y - Z| * \\beta`
    where *Y* is the output and *Z* the expected output,
    :math:`\\alpha` is *l2_weight* and :math:`\\beta`
    is *l1_weight*.

    :param l1_weight: weight of L1 norm
    :param l2_weight: weight of L2 norm

    See @see fn _onnx_grad_loss_elastic_error for the ONNX
    implementation.
    """

    def __init__(self, l1_weight=0.5, l2_weight=0.5):
        BaseLearningLoss.__init__(self)
        self.l1_weight = l1_weight
        self.l2_weight = l2_weight

    def build_onnx_function(self, opset, device, weight_name):
        so = SessionOptions()
        so.log_severity_level = 4

        # loss_grad
        self.loss_grad_onnx_ = function_onnx_graph(
            "grad_loss_elastic_error", target_opset=opset,
            weight_name=weight_name, l1_weight=self.l1_weight,
            l2_weight=self.l2_weight)
        self.loss_grad_sess_ = InferenceSession(
            self.loss_grad_onnx_.SerializeToString(), so,
            providers=device_to_providers(device))
        self.loss_grad_sess_bind_ = (
            self.loss_grad_sess_.io_binding()._iobinding)


class NegLogLearningLoss(BaseLearningLoss):
    """
    Implements a negative log loss
    `'log(yt, yp) = -(1-yt)\\log(1-yp) - yt\\log(yp)`,
    this only works for a binary classification where *yp* is the
    predicted probability, *yt* is the expected probability.
    *yt* is expected to be binary, *yp* is a matrix with two
    columns, the sum on every line is 1.
    However, this loss is usually applied after a function softmax
    and the gradient is directly computed from the loss to the
    raw score before they are processed through the softmax function
    (see class `Log
    <https://github.com/scikit-learn/scikit-learn/blob/main/sklearn/
    linear_model/_sgd_fast.pyx#L236>`_).

    :param eps: clipping value for probabilities,
        avoids computing `log(0)`
    :param probability_function: function to convert
        raw scores into probabilities, default value is `sigmoid`
        for a logistic regression
    """

    def __init__(self, eps=1e-5, probability_function='sigmoid'):
        BaseLearningLoss.__init__(self)
        self.eps = eps
        self.probability_function = probability_function

    def build_onnx_function(self, opset, device, weight_name):
        so = SessionOptions()
        so.log_severity_level = 4

        # loss_grad
        fct_name = "grad_%s_neg_log_loss_error" % self.probability_function
        self.loss_grad_onnx_ = function_onnx_graph(
            fct_name, target_opset=opset,
            weight_name=weight_name, eps=self.eps)
        self.loss_grad_sess_ = InferenceSession(
            self.loss_grad_onnx_.SerializeToString(), so,
            providers=device_to_providers(device))
        self.loss_grad_sess_bind_ = (
            self.loss_grad_sess_.io_binding()._iobinding)



"""
@file
@brief Helpers to display internal structures.
"""


def str_ortvalue(ov):
    """
    Displays the content of an :epkg:`C_OrtValue`.

    :param ov: :epkg:`OrtValue` or :epkg:`C_OrtValue`
    :return: str
    """
    if hasattr(ov, '_ortvalue'):
        return str_ortvalue(ov._ortvalue)
    device = ov.device_name()
    value = ov.numpy()
    values = value.ravel().tolist()
    if len(values) > 10:
        values = values[:5] + ["..."] + values[-5:]
    return "device=%s dtype=%r shape=%r value=%r" % (
        device, value.dtype, value.shape, values)



# -*- coding: utf-8 -*-
import sys
import os
import warnings
from setuptools import setup, Extension, find_packages
from pyquicksetup import read_version, read_readme, default_cmdclass

#########
# settings
#########

project_var_name = "onnxcustom"
versionPython = "%s.%s" % (sys.version_info.major, sys.version_info.minor)
path = "Lib/site-packages/" + project_var_name
readme = 'README.rst'
history = "HISTORY.rst"
requirements = None

KEYWORDS = project_var_name + ', Xavier Dupré'
DESCRIPTION = """Extends scikit-learn with a couple of new models, transformers, metrics, plotting."""
CLASSIFIERS = [
    'Programming Language :: Python :: 3',
    'Intended Audience :: Developers',
    'Topic :: Scientific/Engineering',
    'Topic :: Education',
    'License :: OSI Approved :: MIT License',
    'Development Status :: 5 - Production/Stable'
]


#######
# data
#######

packages = find_packages()
package_dir = {k: os.path.join('.', k.replace(".", "/")) for k in packages}
package_data = {
    project_var_name + ".data": ["*.csv"],
}


# setup

setup(
    name=project_var_name,
    version=read_version(__file__, project_var_name),
    author='Xavier Dupré',
    author_email='xavier.dupre@gmail.com',
    license="MIT",
    url="http://www.xavierdupre.fr/app/%s/helpsphinx/index.html" % project_var_name,
    download_url="https://github.com/sdpython/%s/" % project_var_name,
    description=DESCRIPTION,
    long_description=read_readme(__file__),
    cmdclass=default_cmdclass(),
    keywords=KEYWORDS,
    classifiers=CLASSIFIERS,
    packages=packages,
    package_dir=package_dir,
    package_data=package_data,
    setup_requires=["pyquicksetup"],
    install_requires=["fire", "numpy", "onnx>=1.10.1", "scipy"],
    extras_require={
        'all': ["fire", "numpy", "onnx>=1.10.1",
                "scipy", "pandas_streaming>=0.3", "cxxfilt"]
    }
)
